#!/usr/bin/env python

##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import os

import Image
import ImageChops
import ImageDraw
import ImageOps

def draw_marker(width, height, offset, filename):
    im = Image.new('L', (128, 256), 255)
    draw = ImageDraw.Draw(im)

    cx = (im.size[0] / 2) + 1.5
    cy = (im.size[1] / 4) + 1.5
    radius = width * 4.0 / 2.0
    offset = int(offset * 4.0)
    for thickness, colour in ((0, 0x60), (4, 0xE0)):
        ol = im.copy()
        r = radius + 0.5 - thickness
        draw.ellipse((cx - r, cy - r, cx + r, cy + r), fill=colour)
        x1 = offset + 0.5 - (thickness * 1.5)
        y1 = (height * 4.0) + 0.5 - (width * 2.0) - (thickness * 2.0)
        draw.polygon((cx + 0.5, cy + y1, cx - 0.5, cy + y1,
                      cx - x1, cy + 16.5, cx + x1, cy + 16.5),
                     fill=colour)
    del draw

    im = im.resize((im.size[0] / 4, im.size[1] / 4), Image.ANTIALIAS)
    ol = ol.resize(im.size, Image.ANTIALIAS)
    mask = Image.eval(ol, lambda x: (0, 255)[x < 210])
    im = Image.composite(im, ImageChops.constant(im, 0), mask)
    im = im.crop(im.getbbox())
    im.save(filename, transparency=0)

path = os.path.normpath(os.path.join(
    os.path.dirname(__file__), '..', 'photini', 'data'))
draw_marker(25, 37, 8.5, os.path.join(path, 'bing_grey_marker.png'))
draw_marker(25, 40, 12, os.path.join(path, 'osm_grey_marker.png'))
draw_marker(21.5, 40, 10.0, os.path.join(path, 'google_grey_marker.png'))



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from datetime import datetime
import logging
import os
import sys
import time
import webbrowser

import atom
import gdata.auth
import gdata.photos
import gdata.photos.service
from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from .descriptive import MultiLineEdit
from .utils import Busy

EPOCH = datetime.utcfromtimestamp(0)

logger = logging.getLogger(__name__)

class LoggingPhotosService(object):
    def __init__(self):
        self._service = gdata.photos.service.PhotosService(email='default')

    def __getattr__(self, name):
        self._next_call = getattr(self._service, name)
        return self._call

    def _call(self, *arg, **kw):
        try:
            return (self._next_call)(*arg, **kw)
        except Exception as ex:
            logger.exception(ex)

class MediaSourceWithCallback(object):
    def __init__(self, path, content_type, callback):
        self.file = open(path, 'rb')
        self.content_type = content_type
        self.content_length = os.path.getsize(path)
        self.file_handle = self
        self.callback = callback

    def read(self, size):
        if self.callback:
            self.callback(self.file.tell() * 100 / self.content_length)
        return self.file.read(size)

class UploadThread(QtCore.QThread):
    def __init__(self, pws, upload_list, feed_uri):
        QtCore.QThread.__init__(self)
        self.pws = pws
        self.upload_list = upload_list
        self.feed_uri = feed_uri

    finished = QtCore.pyqtSignal()
    def run(self):
        self.file_count = 0
        for params in self.upload_list:
            photo = gdata.photos.PhotoEntry()
            if not photo.title:
                photo.title = atom.Title()
            photo.title.text = os.path.basename(params['path'])
            if params['summary']:
                if not photo.summary:
                    photo.summary = atom.Summary()
                photo.summary.text = params['summary']
            if params['keywords']:
                if not photo.media:
                    photo.media = gdata.media.Group()
                if not photo.media.keywords:
                    photo.media.keywords = gdata.media.Keywords()
                photo.media.keywords.text = params['keywords']
            if params['latlong']:
                if not photo.geo:
                    photo.geo = gdata.geo.Where()
                if not photo.geo.Point:
                    photo.geo.Point = gdata.geo.Point()
                photo.geo.Point.pos = gdata.geo.Pos(text=params['latlong'])
            mediasource = MediaSourceWithCallback(
                params['path'], 'image/jpeg', self.callback)
            try:
                photo = self.pws.Post(
                    photo, uri=self.feed_uri, media_source=mediasource,
                    converter=gdata.photos.PhotoEntryFromString)
            except gdata.service.RequestError, e:
                raise gdata.photos.service.GooglePhotosException(e.args[0])
            self.file_count += 1
        self.finished.emit()

    progress_report = QtCore.pyqtSignal(float, float)
    def callback(self, progress):
        total_progress = (
            (self.file_count * 100) + progress) / len(self.upload_list)
        self.progress_report.emit(progress, total_progress)

def decode_text(obj):
    if not obj or obj.text is None:
        return ''
    return unicode(obj.text, 'utf-8')

class PicasaUploader(QtGui.QWidget):
    def __init__(self, config_store, image_list, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.image_list = image_list
        self.setLayout(QtGui.QGridLayout())
        self.pws = None
        self.widgets = {}
        self.current_album = None
        self.uploader = None
        ### album group
        self.album_group = QtGui.QGroupBox('Album')
        self.album_group.setLayout(QtGui.QHBoxLayout())
        self.layout().addWidget(self.album_group, 0, 0, 3, 3)
        ## album details, left hand side
        album_form_left = QtGui.QFormLayout()
        album_form_left.setFieldGrowthPolicy(
            QtGui.QFormLayout.AllNonFixedFieldsGrow)
        self.album_group.layout().addLayout(album_form_left)
        # album title / selector
        self.albums = QtGui.QComboBox()
        self.albums.setEditable(True)
        self.albums.setInsertPolicy(QtGui.QComboBox.NoInsert)
        self.albums.currentIndexChanged.connect(self.changed_album)
        self.albums.lineEdit().editingFinished.connect(self.new_title)
        album_form_left.addRow('Title', self.albums)
        # album description
        self.widgets['description'] = MultiLineEdit()
        self.widgets['description'].editingFinished.connect(self.new_description)
        album_form_left.addRow('Description', self.widgets['description'])
        # album location
        self.widgets['location'] = QtGui.QLineEdit()
        self.widgets['location'].editingFinished.connect(self.new_location)
        album_form_left.addRow('Place taken', self.widgets['location'])
        # album visibility
        self.widgets['access'] = QtGui.QComboBox()
        self.widgets['access'].addItem('Public on the web', 'public')
        self.widgets['access'].addItem('Limited, anyone with the link', 'private')
        self.widgets['access'].addItem('Only you', 'protected')
        self.widgets['access'].currentIndexChanged.connect(self.new_access)
        album_form_left.addRow('Visibility', self.widgets['access'])
        ## album buttons
        buttons = QtGui.QHBoxLayout()
        album_form_left.addRow('', buttons)
        # new album
        new_album_button = QtGui.QPushButton('New album')
        new_album_button.clicked.connect(self.new_album)
        buttons.addWidget(new_album_button)
        # delete album
        delete_album_button = QtGui.QPushButton('Delete album')
        delete_album_button.clicked.connect(self.delete_album)
        buttons.addWidget(delete_album_button)
        # other init
        self.clear_changes()
        ## album details, right hand side
        album_form_right = QtGui.QFormLayout()
        album_form_right.setFieldGrowthPolicy(
            QtGui.QFormLayout.AllNonFixedFieldsGrow)
        self.album_group.layout().addLayout(album_form_right)
        # album date
        self.widgets['timestamp'] = QtGui.QDateEdit()
        self.widgets['timestamp'].setCalendarPopup(True)
        self.widgets['timestamp'].editingFinished.connect(self.new_timestamp)
        album_form_right.addRow('Date', self.widgets['timestamp'])
        # album thumbnail
        self.album_thumb = QtGui.QLabel()
        album_form_right.addRow(self.album_thumb)
        ### upload button
        self.upload_button = QtGui.QPushButton('Upload\nnow')
        self.upload_button.setEnabled(False)
        self.upload_button.clicked.connect(self.upload)
        self.layout().addWidget(self.upload_button, 2, 3)
        ### progress bar
        self.layout().addWidget(QtGui.QLabel('Progress'), 3, 0)
        self.total_progress = QtGui.QProgressBar()
        self.layout().addWidget(self.total_progress, 3, 1, 1, 3)
        self.setEnabled(False)
        # adjust spacing
        self.layout().setColumnStretch(2, 1)
        self.layout().setRowStretch(1, 1)
        # timer to store album data after it's edited
        self.timer = QtCore.QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(5000)
        self.timer.timeout.connect(self.save_changes)

    def clear_changes(self):
        self.changed_title = None
        self.changed_description = None
        self.changed_location = None
        self.changed_access = None
        self.changed_timestamp = None

    @QtCore.pyqtSlot()
    def new_title(self):
        value = unicode(self.albums.lineEdit().text())
        self.albums.setItemText(self.albums.currentIndex(), value)
        if value != decode_text(self.current_album.title):
            self.changed_title = value
        else:
            self.changed_title = None
        self.timer.start()

    @QtCore.pyqtSlot()
    def new_description(self):
        value = unicode(self.widgets['description'].text())
        if value != decode_text(self.current_album.summary):
            self.changed_description = value
        else:
            self.changed_description = None
        self.timer.start()

    @QtCore.pyqtSlot()
    def new_timestamp(self):
        value = '%d' % (
            (self.widgets['timestamp'].dateTime().toPyDateTime() - EPOCH
             ).total_seconds() * 1000)
        if value != decode_text(self.current_album.timestamp):
            self.changed_timestamp = value
        else:
            self.changed_timestamp = None
        self.timer.start()

    @QtCore.pyqtSlot()
    def new_location(self):
        value = unicode(self.widgets['location'].text())
        if value != decode_text(self.current_album.location):
            self.changed_location = value
        else:
            self.changed_location = None
        self.timer.start()

    @QtCore.pyqtSlot(int)
    def new_access(self, index):
        value = unicode(self.widgets['access'].itemData(index).toString())
        if value != decode_text(self.current_album.access):
            self.changed_access = value
        else:
            self.changed_access = None
        self.timer.start()

    @QtCore.pyqtSlot()
    def new_album(self):
        self.save_changes()
        with Busy():
            self.current_album = self.pws.InsertAlbum('New album', '')
        self.albums.insertItem(
            0, decode_text(self.current_album.title),
            self.current_album.gphoto_id.text)
        self.albums.setCurrentIndex(0)

    @QtCore.pyqtSlot()
    def delete_album(self):
        if int(self.current_album.numphotos.text) > 0:
            if QtGui.QMessageBox.question(
                self, 'Delete album',
                """Are you sure you want to delete the album "%s"?
Doing so will remove the album and its photos from all Google products.""" % (
    unicode(self.current_album.title.text, 'UTF-8')),
                QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel,
                QtGui.QMessageBox.Cancel
                ) == QtGui.QMessageBox.Cancel:
                return
        self.clear_changes()
        with Busy():
            self.pws.Delete(self.current_album)
        self.albums.removeItem(self.albums.currentIndex())
        if self.albums.count() == 0:
            self.new_album()

    def save_changes(self):
        no_change = True
        # title
        if self.changed_title:
            self.current_album.title.text = self.changed_title.encode('utf-8')
            no_change = False
        # description
        if self.changed_description:
            if not self.current_album.summary:
                self.current_album.summary.text = atom.summary()
            self.current_album.summary.text = self.changed_description.encode('utf-8')
            no_change = False
        # location
        if self.changed_location:
            if not self.current_album.location:
                self.current_album.location = gdata.photos.Location()
            self.current_album.location.text = self.changed_location.encode('utf-8')
            no_change = False
        # access
        if self.changed_access:
            self.current_album.access.text = self.changed_access.encode('utf-8')
            no_change = False
        # timestamp
        if self.changed_timestamp:
            self.current_album.timestamp.text = self.changed_timestamp.encode('utf-8')
            no_change = False
        # upload changes
        self.clear_changes()
        if no_change:
            return
        with Busy():
            self.current_album = self.pws.Put(
                self.current_album, self.current_album.GetEditLink().href,
                converter=gdata.photos.AlbumEntryFromString)

    def refresh(self):
        if self.pws:
            return
        with Busy():
            if not self.authorise():
                return
            albums = self.pws.GetUserFeed()
            for album in albums.entry:
                if not album.GetEditLink():
                    # ignore 'system' albums
                    continue
                self.albums.addItem(
                    unicode(album.title.text, 'UTF-8'), album.gphoto_id.text)
            if self.albums.count() == 0:
                self.new_album()
        self.setEnabled(True)

    @QtCore.pyqtSlot(int)
    def changed_album(self, index):
        self.save_changes()
        self.current_album = None
        album_id = unicode(self.albums.itemData(index).toString())
        with Busy():
            albums = self.pws.GetUserFeed()
            for album in albums.entry:
                if album.gphoto_id.text == album_id:
                    self.current_album = album
                    break
            else:
                return
            if self.current_album.media.thumbnail:
                media = self.pws.GetMedia(self.current_album.media.thumbnail[0].url)
                image = QtGui.QPixmap()
                image.loadFromData(media.file_handle.read())
                self.album_thumb.setPixmap(image)
            else:
                self.album_thumb.clear()
        self.widgets['timestamp'].setDateTime(datetime.utcfromtimestamp(
            float(self.current_album.timestamp.text) * 1.0e-3))
        value = decode_text(self.current_album.summary)
        if value:
            self.widgets['description'].setText(value)
        else:
            self.widgets['description'].clear()
        value = decode_text(self.current_album.location)
        if value:
            self.widgets['location'].setText(value)
        else:
            self.widgets['location'].clear()
        self.widgets['access'].setCurrentIndex(
            self.widgets['access'].findData(self.current_album.access.text))

    @QtCore.pyqtSlot()
    def upload(self):
        if not self.image_list.unsaved_files_dialog(with_discard=False):
            return
        # make list of items to upload
        upload_list = []
        for image in self.image_list.get_selected_images():
            title = image.metadata.get_item('title').as_str()
            description = image.metadata.get_item('description').as_str()
            if title and description:
                summary = '%s\n\n%s' % (title, description)
            elif title:
                summary = title
            else:
                summary = description
            keywords = image.metadata.get_item('keywords')
            if keywords.empty():
                keywords = None
            else:
                keywords = ', '.join(keywords.value)
            latlong = image.metadata.get_item('latlong')
            if latlong.empty():
                latlong = None
            else:
                latlong = '%f %f' % latlong.value
            upload_list.append({
                'path'     : image.path,
                'summary'  : summary,
                'keywords' : keywords,
                'latlong'  : latlong,
                })
        # pass the list to a separate thread, so GUI can continue
        if self.authorise():
            self.upload_button.setEnabled(False)
            self.album_group.setEnabled(False)
            self.uploader = UploadThread(
                self.pws, upload_list, self.current_album.GetFeedLink().href)
            self.uploader.progress_report.connect(self.upload_progress)
            self.uploader.finished.connect(self.upload_done)
            self.uploader.start()
            # we've passed the service handle to a new thread, so
            # create a new one for safety
            self.pws = None
            self.authorise()

    @QtCore.pyqtSlot(float, float)
    def upload_progress(self, progress, total_progress):
        self.total_progress.setValue(total_progress)

    @QtCore.pyqtSlot()
    def upload_done(self):
        # reload current album
        self.changed_album(self.albums.currentIndex())
        self.upload_button.setEnabled(True)
        self.album_group.setEnabled(True)
        self.total_progress.setValue(0)
        self.uploader = None

    def authorise(self):
        if self.pws:
            return True
        self.pws = LoggingPhotosService()
        self.pws.SetOAuthInputParameters(
            gdata.auth.OAuthSignatureMethod.HMAC_SHA1,
            b'991146392375.apps.googleusercontent.com',
            consumer_secret=b'gSCBPBV0tpArWOK2IDcEA6eG')
        key = self.config_store.get('picasa', 'key', '')
        secret = self.config_store.get('picasa', 'secret', '')
        if sys.version_info[0] < 3:
            key = str(key)
            secret = str(secret)
        if key and secret:
            token = gdata.auth.OAuthToken(
                key=key, secret=secret,
                scopes='https://picasaweb.google.com/data/',
                oauth_input_params=self.pws.GetOAuthInputParameters())
            self.pws.SetOAuthToken(token)
            return True
        request_token = self.pws.FetchOAuthRequestToken(
            scopes='https://picasaweb.google.com/data/',
            extra_parameters={'xoauth_displayname' : 'Photini'},
            oauth_callback='oob')
        self.pws.SetOAuthToken(request_token)
        auth_url = self.pws.GenerateOAuthAuthorizationURL()
        if webbrowser.open(auth_url, new=2, autoraise=0):
            info_text = 'use your web browser'
        else:
            info_text = 'open "%s" in a web browser' % auth_url
        auth_code, OK = QtGui.QInputDialog.getText(
            self,
            'Photini: authorise Picasa',
            """Please %s to grant access to Photini,
then enter the verification code:""" % info_text)
        if not OK:
            self.pws = None
            return False
        try:
            access_token = self.pws.UpgradeToOAuthAccessToken(
                oauth_verifier=str(auth_code))
        except gdata.service.TokenUpgradeFailed:
            self.pws = None
            return False
        self.config_store.set('picasa', 'key', access_token.key)
        self.config_store.set('picasa', 'secret', access_token.secret)
        return True

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        self.upload_button.setEnabled(
            len(selection) > 0 and self.pws and not self.uploader)



#!/usr/bin/env python

##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from datetime import datetime
import logging
import os
import re
import sys

import gphoto2 as gp
from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from .configstore import ConfigStore
from .utils import Busy

class CameraHandler(QtCore.QObject):
    new_camera_list = QtCore.pyqtSignal(list)
    new_camera = QtCore.pyqtSignal(str)
    def __init__(self, parent=None):
        QtCore.QObject.__init__(self, parent)
        self.context = gp.Context()
        self.camera = None
        self.cam_model = ''
        self.cam_port_name = None

    @QtCore.pyqtSlot()
    def get_camera_list(self):
        with gp.CameraList() as cameras:
            cam_count = self.context.camera_autodetect(cameras.list)
            camera_list = []
            for n in range(cam_count):
                name = cameras.get_name(n)
                addr = cameras.get_value(n)
                camera_list.append((name, addr))
        camera_list.sort(key=lambda x: x[0])
        self.new_camera_list.emit(camera_list)

    @QtCore.pyqtSlot(str, str)
    def select_camera(self, model, port_name):
        # free any existing camera
        if self.camera:
            self.camera.exit()
            self.camera = None
        # initialise camera
        self.camera = gp.Camera(self.context.context)
        # search abilities for camera model
        with gp.CameraAbilitiesList() as abilities_list:
            abilities_list.load(self.context.context)
            idx = abilities_list.lookup_model(str(model))
            abilities = gp.CameraAbilities()
            abilities_list.get_abilities(idx, abilities)
            self.camera.set_abilities(abilities)
        # search ports for camera port name
        with gp.PortInfoList() as port_info_list:
            port_info_list.load()
            idx = port_info_list.lookup_path(str(port_name))
            # port_info is a pointer to an entry in port_info_list, so
            # don't free port_info_list until after port_info has been
            # used
            port_info = port_info_list.get_info(idx)
            self.camera.set_port_info(port_info)
        self.cam_model = model
        self.cam_port_name = port_name
        try:
            self.camera.init()
        except gp.GPhoto2Error:
            self.camera = None
            self.cam_model = ''
            self.cam_port_name = None
            self.get_camera_list()
        self.new_camera.emit(self.cam_model)

    def list_files(self, path='/'):
        if not self.camera:
            return []
        result = []
        with gp.CameraList() as gp_list:
            # get files
            self.camera.folder_list_files(str(path), gp_list.list)
            for n in range(gp_list.count()):
                result.append(os.path.join(path, gp_list.get_name(n)))
            # read folders
            folders = []
            gp_list.reset()
            self.camera.folder_list_folders(str(path), gp_list.list)
            for n in range(gp_list.count()):
                folders.append(gp_list.get_name(n))
        # recurse over subfolders
        for name in folders:
            result.extend(self.list_files(os.path.join(path, name)))
        return result

    def get_file_info(self, path):
        folder, name = os.path.split(path)
        info = gp.CameraFileInfo()
        self.camera.file_get_info(str(folder), str(name), info)
        return info

    def copy_file(self, folder, name, dest):
        with gp.CameraFile() as camera_file:
            self.camera.file_get(
                str(folder), str(name), gp.GP_FILE_TYPE_NORMAL, camera_file.file)
            camera_file.save(dest)

class CameraSelector(QtGui.QWidget):
    select_camera = QtCore.pyqtSignal(str, str)
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.setLayout(QtGui.QHBoxLayout())
        self.layout().setMargin(0)
        # camera selector
        self.cam_list = QtGui.QComboBox()
        self.cam_list.activated.connect(self._activated)
        self.layout().addWidget(self.cam_list)
        self.current_cam = ''
        # rescan button
        scan_button = QtGui.QPushButton('rescan')
        self.scan_for_cameras = scan_button.clicked
        self.layout().addWidget(scan_button)
        # make drop down list as wide as possible
        self.layout().setStretch(0, 1)

    @QtCore.pyqtSlot(str)
    def camera_changed(self, model):
        self.current_cam = model
        self._show_current()

    @QtCore.pyqtSlot(list)
    def new_camera_list(self, cameras):
        first_time = self.cam_list.count() == 0
        self.cam_list.setUpdatesEnabled(False)
        self.cam_list.clear()
        self.cam_list.addItem('<select camera>', None)
        for model, port_name in cameras:
            self.cam_list.addItem(model, port_name)
        if first_time and len(cameras) == 1:
            self.select_camera.emit(*cameras[0])
        else:
            self._show_current()
        self.cam_list.setUpdatesEnabled(True)

    @QtCore.pyqtSlot(int)
    def _activated(self, idx):
        port_name = self.cam_list.itemData(idx)
        if sys.version_info[0] < 3:
            port_name = port_name.toString()
        if port_name:
            model = self.cam_list.itemText(idx)
            if model != self.current_cam:
                self.select_camera.emit(model, port_name)
                return
        self._show_current()

    def _show_current(self):
        self.cam_list.setCurrentIndex(
            max(0, self.cam_list.findText(self.current_cam)))

class NameMangler(QtCore.QObject):
    number_parser = re.compile('\D*(\d+)')
    new_example = QtCore.pyqtSignal(str)
    def __init__(self, parent=None):
        QtCore.QObject.__init__(self, parent)
        self.example = None
        self.format_string = None

    @QtCore.pyqtSlot(str)
    def new_format(self, format_string):
        format_string = str(format_string)
        self.format_string = format_string
        # extract bracket delimited words from string
        self.parts = []
        while format_string:
            parts = format_string.split('(', 1)
            if len(parts) > 1:
                parts[1:] = parts[1].split(')', 1)
            if len(parts) < 3:
                self.parts.append((format_string, ''))
                break
            self.parts.append((parts[0], parts[1]))
            format_string = parts[2]
        self.refresh_example()

    def set_example(self, name, timestamp):
        self.example = name, timestamp
        self.refresh_example()

    def refresh_example(self):
        if self.format_string and self.example:
            self.new_example.emit(self.transform(*self.example))

    def transform(self, name, timestamp):
        subst = {'name': name}
        match = self.number_parser.match(name)
        if match:
            subst['number'] = match.group(1)
        else:
            subst['number'] = ''
        subst['root'], subst['ext'] = os.path.splitext(name)
        result = ''
        # process (...) parts first
        for left, right in self.parts:
            result += left
            if right in subst:
                result += subst[right]
            else:
                result += right
        # then do timestamp
        return timestamp.strftime(result)

class Importer(QtGui.QWidget):
    def __init__(self, config_store, image_list, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.image_list = image_list
        self.setLayout(QtGui.QGridLayout())
        form = QtGui.QFormLayout()
        form.setFieldGrowthPolicy(QtGui.QFormLayout.AllNonFixedFieldsGrow)
        self.app = QtGui.QApplication.instance()
        self.ch = CameraHandler()
        self.nm = NameMangler()
        self.camera = ''
        self.file_data = {}
        self.file_list = []
        self.config_section = 'importer'
        # camera selector
        self.camera_selector = CameraSelector()
        self.camera_selector.scan_for_cameras.connect(self.ch.get_camera_list)
        self.camera_selector.select_camera.connect(self.ch.select_camera)
        self.ch.new_camera_list.connect(self.camera_selector.new_camera_list)
        self.ch.new_camera.connect(self.new_camera)
        form.addRow('Camera', self.camera_selector)
        # path format
        self.path_format = QtGui.QLineEdit()
        self.path_format.textChanged.connect(self.nm.new_format)
        self.path_format.editingFinished.connect(self.path_format_finished)
        form.addRow('Target format', self.path_format)
        # path example
        self.path_example = QtGui.QLabel()
        self.nm.new_example.connect(self.path_example.setText)
        form.addRow('=>', self.path_example)
        self.layout().addLayout(form, 0, 0)
        # file list
        self.file_list_widget = QtGui.QListWidget()
        self.file_list_widget.setSelectionMode(
            QtGui.QAbstractItemView.ExtendedSelection)
        self.layout().addWidget(self.file_list_widget, 1, 0)
        # selection buttons
        buttons = QtGui.QVBoxLayout()
        buttons.addStretch(1)
        select_all = QtGui.QPushButton('Select\nall')
        select_all.clicked.connect(self.select_all)
        buttons.addWidget(select_all)
        select_new = QtGui.QPushButton('Select\nnew')
        select_new.clicked.connect(self.select_new)
        buttons.addWidget(select_new)
        copy_selected = QtGui.QPushButton('Copy\nphotos')
        copy_selected.clicked.connect(self.copy_selected)
        buttons.addWidget(copy_selected)
        self.layout().addLayout(buttons, 0, 1, 2, 1)
        # final initialisation
        self.image_list.sort_order_changed.connect(self.sort_file_list)
        self.path_format.setText(os.path.join(
            os.path.expanduser('~/Pictures'), '%Y/%Y_%m_%d/(name)'))
        self.new_camera('')

    @QtCore.pyqtSlot()
    def path_format_finished(self):
        if self.camera:
            self.config_store.set(
                self.config_section, 'path_format', self.nm.format_string)
        self.show_file_list()

    def refresh(self):
        self.ch.get_camera_list()

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        pass

    @QtCore.pyqtSlot(str)
    def new_camera(self, camera_model):
        self.camera = unicode(camera_model)
        if self.camera:
            self.config_section = 'importer %s' % self.camera
            path_format = unicode(self.path_format.text())
            path_format = self.config_store.get(
                self.config_section, 'path_format', path_format)
            self.path_format.setText(path_format)
        self.camera_selector.camera_changed(camera_model)
##        self.statusBar().showMessage('Getting file list...')
        self.file_list_widget.clear()
        # allow 100ms for display to update before getting file list
        QtCore.QTimer.singleShot(100, self.list_files)

    def list_files(self):
        self.file_data = {}
        with Busy():
            file_list = []
            try:
                file_list = self.ch.list_files()
            except Exception:
                pass
            for path in file_list:
                folder, name = os.path.split(path)
                info = self.ch.get_file_info(path)
                timestamp = datetime.fromtimestamp(info.file.mtime)
                self.file_data[name] = {
                    'src_folder' : folder,
                    'timestamp'  : timestamp,
                    }
##        self.statusBar().clearMessage()
        self.file_list = self.file_data.keys()
        self.sort_file_list()

    def sort_file_list(self):
        if eval(self.config_store.get('controls', 'sort_date', 'False')):
            self.file_list.sort(key=lambda x: self.file_data[x]['timestamp'])
        else:
            self.file_list.sort()
        self.show_file_list()
        if self.file_list:
            name = self.file_list[-1]
            timestamp = self.file_data[name]['timestamp']
        else:
            name = 'IMG_9999.JPG'
            timestamp = datetime.now()
        self.nm.set_example(name, timestamp)

    def show_file_list(self):
        self.file_list_widget.clear()
        first_active = None
        item = None
        for name in self.file_list:
            timestamp = self.file_data[name]['timestamp']
            dest_path = self.nm.transform(name, timestamp)
            self.file_data[name]['dest_path'] = dest_path
            item = QtGui.QListWidgetItem('%s -> %s' % (name, dest_path))
            if os.path.exists(dest_path):
                item.setFlags(Qt.NoItemFlags)
            else:
                if not first_active:
                    first_active = item
                item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            self.file_list_widget.addItem(item)
        if not first_active:
            first_active = item
        self.file_list_widget.scrollToItem(
            first_active, QtGui.QAbstractItemView.PositionAtTop)

    @QtCore.pyqtSlot()
    def select_all(self):
        self.select_files(datetime.min)

    @QtCore.pyqtSlot()
    def select_new(self):
        since = datetime.min
        if self.camera:
            since = datetime.strptime(
                self.config_store.get(
                    self.config_section, 'last_transfer', since.isoformat(b' ')),
                '%Y-%m-%d %H:%M:%S')
        self.select_files(since)

    def select_files(self, since):
        count = self.file_list_widget.count()
        if not count:
            return
        self.file_list_widget.clearSelection()
        first_active = None
        for row in range(count):
            item = self.file_list_widget.item(row)
            if not (item.flags() & Qt.ItemIsSelectable):
                continue
            name = str(item.text()).split()[0]
            timestamp = self.file_data[name]['timestamp']
            if timestamp > since:
                if not first_active:
                    first_active = item
                self.file_list_widget.setItemSelected(item, True)
        if not first_active:
            first_active = item
        self.file_list_widget.scrollToItem(
            first_active, QtGui.QAbstractItemView.PositionAtTop)

    @QtCore.pyqtSlot()
    def copy_selected(self):
        indexes = self.file_list_widget.selectedIndexes()
        if not indexes:
            return
        last_transfer = datetime.min
        count = 0
        with Busy():
            for idx in indexes:
                count += 1
                item = self.file_list_widget.itemFromIndex(idx)
                name = str(item.text()).split()[0]
                timestamp = self.file_data[name]['timestamp']
                dest_path = self.file_data[name]['dest_path']
                src_folder = self.file_data[name]['src_folder']
##                self.statusBar().showMessage(
##                    'Copying %d/%d %s' % (count, len(indexes), dest_path))
                self.app.processEvents()
                dest_dir = os.path.dirname(dest_path)
                if not os.path.isdir(dest_dir):
                    os.makedirs(dest_dir)
                self.ch.copy_file(src_folder, name, dest_path)
                self.image_list.open_file_list([dest_path])
                last_transfer = max(last_transfer, timestamp)
##        self.statusBar().clearMessage()
        self.show_file_list()
        self.config_store.set(
            self.config_section, 'last_transfer', last_transfer.isoformat(b' '))



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from datetime import (
    timedelta, datetime as pyDateTime, date as pyDate, time as pyTime)

from PyQt4 import QtGui, QtCore

class DateEdit(QtGui.QDateEdit):
    def focusInEvent(self, event):
        if self.date() == self.minimumDate():
            self.setDate(QtCore.QDate.currentDate())
        return QtGui.QDateEdit.focusInEvent(self, event)

class TimeEdit(QtGui.QTimeEdit):
    def focusInEvent(self, event):
        if self.time() == self.minimumTime():
            self.setSpecialValueText('')
        return QtGui.QTimeEdit.focusInEvent(self, event)

class DateAndTimeWidget(QtGui.QWidget):
    def __init__(self, key, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.key = key
        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        # date
        self.date = DateEdit()
        self.date.setDisplayFormat('yyyy-MM-dd')
        self.date.setCalendarPopup(True)
        self.date.setSpecialValueText(' ')
        self.date.editingFinished.connect(self._new_date)
        layout.addWidget(self.date)
        # time
        self.time = TimeEdit()
        self.time.setDisplayFormat('hh:mm:ss')
        self.time.setSpecialValueText(' ')
        self.time.editingFinished.connect(self._new_time)
        layout.addWidget(self.time)
        # clear button
        clear_button = QtGui.QPushButton('clear')
        clear_button.clicked.connect(self._clear)
        layout.addWidget(clear_button)

    date_changed = QtCore.pyqtSignal(str, QtCore.QDate)
    def _new_date(self):
        date = self.date.date()
        self.date_changed.emit(self.key, date)

    time_changed = QtCore.pyqtSignal(str, QtCore.QTime)
    def _new_time(self):
        time = self.time.time()
        self.time_changed.emit(self.key, time)

    def _clear(self):
        self.clearDate()
        self.clearTime()
        self.time_changed.emit(self.key, self.time.minimumTime())
        self.date_changed.emit(self.key, self.date.minimumDate())

    def clearDate(self):
        self.date.setSpecialValueText(' ')
        self.date.setDate(self.date.minimumDate())

    def clearTime(self):
        self.time.setSpecialValueText(' ')
        self.time.setTime(self.time.minimumTime())

    def setMultipleDate(self):
        self.date.setSpecialValueText('<multiple>')
        self.date.setDate(self.date.minimumDate())

    def setMultipleTime(self):
        self.time.setSpecialValueText('<multi>')
        self.time.setTime(self.time.minimumTime())

    def setDate(self, value):
        self.date.setDate(value)

    def setTime(self, value):
        self.time.setTime(value)
        self.time.setSpecialValueText('')

class OffsetWidget(QtGui.QWidget):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        # hours
        self.hours = QtGui.QSpinBox()
        self.hours.setRange(-12, 12)
        self.hours.setSuffix(' hours')
        layout.addWidget(self.hours)
        # minutes
        self.minutes = QtGui.QSpinBox()
        self.minutes.setRange(-60, 60)
        self.minutes.setSuffix(' minutes')
        layout.addWidget(self.minutes)
        # apply button
        apply_button = QtGui.QPushButton('apply')
        apply_button.clicked.connect(self._apply)
        layout.addWidget(apply_button)

    apply_offset = QtCore.pyqtSignal(timedelta)
    def _apply(self):
        offset = timedelta(hours=self.hours.value(),
                           minutes=self.minutes.value())
        self.apply_offset.emit(offset)

class Technical(QtGui.QWidget):
    def __init__(self, config_store, image_list, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.image_list = image_list
        self.setLayout(QtGui.QGridLayout())
        self.date_widget = {}
        self.link_widget = {}
        self.link_master = {
            'taken'     : None,
            'digitised' : 'taken',
            'modified'  : 'digitised'
            }
        # date and time
        date_group = QtGui.QGroupBox('Date and time')
        date_group.setLayout(QtGui.QFormLayout())
        # taken
        self.date_widget['taken'] = DateAndTimeWidget('taken')
        self.date_widget['taken'].date_changed.connect(self.new_datetime)
        self.date_widget['taken'].time_changed.connect(self.new_datetime)
        date_group.layout().addRow('Taken', self.date_widget['taken'])
        # link taken & digitised
        self.link_widget['digitised'] = QtGui.QCheckBox(
            "Link 'taken' and 'digitised'")
        self.link_widget['digitised'].clicked.connect(self.new_link)
        date_group.layout().addRow('', self.link_widget['digitised'])
        # digitised
        self.date_widget['digitised'] = DateAndTimeWidget('digitised')
        self.date_widget['digitised'].date_changed.connect(self.new_datetime)
        self.date_widget['digitised'].time_changed.connect(self.new_datetime)
        date_group.layout().addRow('Digitised', self.date_widget['digitised'])
        # link digitised & modified
        self.link_widget['modified'] = QtGui.QCheckBox(
            "Link 'digitised' and 'modified'")
        self.link_widget['modified'].clicked.connect(self.new_link)
        date_group.layout().addRow('', self.link_widget['modified'])
        # modified
        self.date_widget['modified'] = DateAndTimeWidget('modified')
        self.date_widget['modified'].date_changed.connect(self.new_datetime)
        self.date_widget['modified'].time_changed.connect(self.new_datetime)
        date_group.layout().addRow('Modified', self.date_widget['modified'])
        # offset
        self.offset_widget = OffsetWidget()
        self.offset_widget.apply_offset.connect(self.apply_offset)
        date_group.layout().addRow('Offset', self.offset_widget)
        self.layout().addWidget(date_group, 0, 0)
        # other
        other_group = QtGui.QGroupBox('Other')
        other_group.setLayout(QtGui.QFormLayout())
        # orientation
        self.orientation = QtGui.QComboBox()
        self.orientation.addItem('normal', 1)
        self.orientation.addItem('rotate -90', 6)
        self.orientation.addItem('rotate +90', 8)
        self.orientation.addItem('rotate 180', 3)
        self.orientation.addItem('reflect left-right', 2)
        self.orientation.addItem('reflect top-bottom', 4)
        self.orientation.addItem('reflect tr-bl', 5)
        self.orientation.addItem('reflect tl-br', 7)
        self.orientation.addItem('<multiple>', -1)
        self.orientation.currentIndexChanged.connect(self.new_orientation)
        other_group.layout().addRow('Orientation', self.orientation)
        self.layout().addWidget(other_group, 0, 1)
        # disable until an image is selected
        for key in self.date_widget:
            self.date_widget[key].setEnabled(False)
        self.offset_widget.setEnabled(False)
        self.orientation.setEnabled(False)

    def refresh(self):
        pass

    def new_datetime(self, key, value):
        key = str(key)
        self._new_datetime_value(key, value)
        for slave in self.link_widget:
            if self.link_master[slave] == key:
                if self.link_widget[slave].isChecked():
                    self.new_datetime(slave, value)
                break

    def apply_offset(self, offset):
        for image in self.image_list.get_selected_images():
            for key in self.date_widget:
                value = image.metadata.get_item('date_%s' % key)
                if value.empty():
                    continue
                image.metadata.set_item('date_%s' % key, value.value + offset)
        for key in self.date_widget:
            self._update_datetime(key)

    def new_link(self):
        for key in self.link_widget:
            master = self.link_master[key]
            if self.link_widget[key].isChecked():
                value = self.date_widget[master].date.date()
                self._new_datetime_value(
                    key, self.date_widget[master].time.time())
                self._new_datetime_value(key, value)
                self.date_widget[key].setEnabled(False)
            else:
                self.date_widget[key].setEnabled(True)

    @QtCore.pyqtSlot(int)
    def new_orientation(self, index):
        value, OK = self.orientation.itemData(index).toInt()
        if value < 1 or not OK:
            self._update_orientation()
            return
        for image in self.image_list.get_selected_images():
            image.metadata.set_item('orientation', value)
            image.pixmap = None
            image.load_thumbnail()

    def _new_datetime_value(self, key, value):
        if isinstance(value, QtCore.QTime):
            # update times, leaving date unchanged
            for image in self.image_list.get_selected_images():
                current = image.metadata.get_item('date_%s' % key)
                if current.empty():
                    current = pyDateTime.today()
                else:
                    current = current.value
                image.metadata.set_item(
                    'date_%s' % key,
                    pyDateTime.combine(current.date(), value.toPyTime()))
        elif value == self.date_widget[key].date.minimumDate():
            # clear date & time
            for image in self.image_list.get_selected_images():
                image.metadata.del_item('date_%s' % key)
        else:
            # update dates, leaving times unchanged
            for image in self.image_list.get_selected_images():
                current = image.metadata.get_item('date_%s' % key)
                if current.empty():
                    current = pyDateTime.min
                else:
                    current = current.value
                image.metadata.set_item(
                    'date_%s' % key,
                    pyDateTime.combine(value.toPyDate(), current.time()))
        self._update_datetime(key)

    def _update_datetime(self, key):
        dates = []
        times = []
        for image in self.image_list.get_selected_images():
            value = image.metadata.get_item('date_%s' % key)
            if value.empty():
                dates.append(None)
                times.append(None)
            else:
                dates.append(value.value.date())
                times.append(value.value.time())
        value = dates[0]
        for new_value in dates[1:]:
            if new_value != value:
                self.date_widget[key].setMultipleDate()
                break
        else:
            if value is None:
                self.date_widget[key].clearDate()
            else:
                self.date_widget[key].setDate(value)
        value = times[0]
        for new_value in times[1:]:
            if new_value != value:
                self.date_widget[key].setMultipleTime()
                break
        else:
            if value is None:
                self.date_widget[key].clearTime()
            else:
                self.date_widget[key].setTime(value)

    def _update_orientation(self):
        value = None
        for image in self.image_list.get_selected_images():
            new_value = image.metadata.get_item('orientation')
            if new_value.empty():
                new_value = None
            else:
                new_value = new_value.value
            if value and new_value != value:
                value = -1
                break
            value = new_value
        if not value:
            value = 1
        self.orientation.setCurrentIndex(self.orientation.findData(value))

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        if not selection:
            for key in self.date_widget:
                self.date_widget[key].clearDate()
                self.date_widget[key].clearTime()
                self.date_widget[key].setEnabled(False)
            for key in self.link_widget:
                self.link_widget[key].setEnabled(False)
            self.offset_widget.setEnabled(False)
            self.orientation.setCurrentIndex(self.orientation.findData(1))
            self.orientation.setEnabled(False)
            return
        self.date_widget['taken'].setEnabled(True)
        for key in self.date_widget:
            self._update_datetime(key)
        for key in self.link_widget:
            master = self.link_master[key]
            self.link_widget[key].setEnabled(True)
            if (self.date_widget[key].date.date() ==
                                        self.date_widget[master].date.date() and
                    self.date_widget[key].time.time() ==
                                        self.date_widget[master].time.time()):
                self.date_widget[key].setEnabled(False)
                self.link_widget[key].setChecked(True)
            else:
                self.link_widget[key].setChecked(False)
                self.date_widget[key].setEnabled(True)
        self.offset_widget.setEnabled(True)
        self.orientation.setEnabled(True)
        self._update_orientation()



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os

from PyQt4 import QtGui
from PyQt4.QtCore import Qt

data_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', '')

class Busy(object):
    def __enter__(self):
        QtGui.QApplication.setOverrideCursor(Qt.WaitCursor)
        return self

    def __exit__(self, type, value, traceback):
        QtGui.QApplication.restoreOverrideCursor()



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-13  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import logging

import pyexiv2

class MetadataHandler(object):
    def __init__(self, path):
        self.logger = logging.getLogger(self.__class__.__name__)
        self._md = pyexiv2.ImageMetadata(path)
        self._md.read()

    def save(self):
        try:
            self._md.write()
        except IOError as ex:
            self.logger.exception(ex)
            return False
        return True

    def copy(self, other, exif=True, iptc=True, xmp=True, comment=True):
        # copy from other to self (pyexiv2 copies from self to other)
        other._md.copy(self._md, exif=exif, iptc=iptc, xmp=xmp, comment=comment)

    def get_tags(self):
        return self.get_exif_tags() + self.get_iptc_tags() + self.get_xmp_tags()

    def get_exif_tags(self):
        return self._md.exif_keys

    def get_iptc_tags(self):
        return self._md.iptc_keys

    def get_xmp_tags(self):
        return self._md.xmp_keys

    def get_tag_string(self, tag):
        return self._md[tag].raw_value

    def get_tag_multiple(self, tag):
        item = self._md[tag]
        if item.type == 'Lang Alt':
            return item.value.values()
        elif item.type == 'Date':
            return [item.raw_value]
        return item.value

    def set_tag_string(self, tag, value):
        if tag.startswith('Exif'):
            new_tag = pyexiv2.ExifTag(tag)
        new_tag.raw_value = value
        self._md[tag] = new_tag

    def set_tag_multiple(self, tag, value):
        if tag.startswith('Iptc'):
            new_tag = pyexiv2.IptcTag(tag, value)
        elif tag.startswith('Xmp'):
            new_tag = pyexiv2.XmpTag(tag)
            if new_tag.type == 'Lang Alt':
                new_tag = pyexiv2.XmpTag(tag, {'x-default': value[0]})
            elif new_tag.type == 'Date':
                new_tag.raw_value = value[0]
            else:
                new_tag = pyexiv2.XmpTag(tag, value)
        self._md[tag] = new_tag

    def clear_tag(self, tag):
        if tag in self._md:
            del self._md[tag]



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from datetime import datetime

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

class MultiLineEdit(QtGui.QPlainTextEdit):
    def __init__(self, parent=None):
        QtGui.QPlainTextEdit.__init__(self, parent)
        self.setText = self.setPlainText
        self.text = self.toPlainText

    editingFinished = QtCore.pyqtSignal()
    def focusOutEvent(self, event):
        self.editingFinished.emit()
        QtGui.QPlainTextEdit.focusOutEvent(self, event)

class LineEditWithAuto(QtGui.QWidget):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        # line edit box
        self.edit = QtGui.QLineEdit()
        layout.addWidget(self.edit)
        # auto complete button
        self.auto = QtGui.QPushButton('Auto')
        layout.addWidget(self.auto)
        # adopt child widget methods and signals
        self.editingFinished = self.edit.editingFinished
        self.setText = self.edit.setText
        self.clear = self.edit.clear
        self.text = self.edit.text
        self.autoComplete = self.auto.clicked

class Descriptive(QtGui.QWidget):
    def __init__(self, config_store, image_list, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.image_list = image_list
        self.form = QtGui.QFormLayout()
        self.setLayout(self.form)
        # construct widgets
        self.widgets = dict()
        # title
        self.widgets['title'] = QtGui.QLineEdit()
        self.widgets['title'].editingFinished.connect(self.new_title)
        self.form.addRow('Title / Object Name', self.widgets['title'])
        # description
        self.widgets['description'] = MultiLineEdit()
        self.widgets['description'].editingFinished.connect(self.new_description)
        self.form.addRow('Description / Caption', self.widgets['description'])
        # keywords
        self.widgets['keywords'] = QtGui.QLineEdit()
        self.widgets['keywords'].editingFinished.connect(self.new_keywords)
        self.form.addRow('Keywords', self.widgets['keywords'])
        # copyright
        self.widgets['copyright'] = LineEditWithAuto()
        self.widgets['copyright'].editingFinished.connect(self.new_copyright)
        self.widgets['copyright'].autoComplete.connect(self.auto_copyright)
        self.form.addRow('Copyright', self.widgets['copyright'])
        # creator
        self.widgets['creator'] = LineEditWithAuto()
        self.widgets['creator'].editingFinished.connect(self.new_creator)
        self.widgets['creator'].autoComplete.connect(self.auto_creator)
        self.form.addRow('Creator / Artist', self.widgets['creator'])
        # disable until an image is selected
        for key in self.widgets:
            self.widgets[key].setEnabled(False)

    def refresh(self):
        pass

    def new_title(self):
        self._new_value('title')

    def new_description(self):
        self._new_value('description')

    def new_keywords(self):
        self._new_value('keywords')

    def new_copyright(self):
        self._new_value('copyright')

    def new_creator(self):
        self._new_value('creator')

    def auto_copyright(self):
        name = self.config_store.get('user', 'copyright_name')
        if not name:
            name, OK = QtGui.QInputDialog.getText(
                self, 'Photini: input name',
                "Please type in the copyright holder's name",
                text=self.config_store.get('user', 'creator_name', ''))
            if OK and name:
                name = unicode(name)
                self.config_store.set('user', 'copyright_name', name)
            else:
                name = ''
        for image in self.image_list.get_selected_images():
            date = image.metadata.get_item('date_taken')
            if date.empty():
                date = datetime.now()
            else:
                date = date.value
            value = 'Copyright ©%d %s. All rights reserved.' % (
                date.year, name)
            image.metadata.set_item('copyright', value)
        self._update_widget('copyright')

    def auto_creator(self):
        name = self.config_store.get('user', 'creator_name')
        if not name:
            name, OK = QtGui.QInputDialog.getText(
                self, 'Photini: input name',
                "Please type in the creator's name",
                text=self.config_store.get('user', 'copyright_name', ''))
            if OK and name:
                name = unicode(name)
                self.config_store.set('user', 'creator_name', name)
            else:
                name = ''
        self.widgets['creator'].setText(name)
        self._new_value('creator')

    def _new_value(self, key):
        value = unicode(self.widgets[key].text())
        if value != '<multiple values>':
            for image in self.image_list.get_selected_images():
                image.metadata.set_item(key, value)
        self._update_widget(key)

    def _update_widget(self, key):
        value = None
        for image in self.image_list.get_selected_images():
            new_value = image.metadata.get_item(key).as_str()
            if value and new_value != value:
                self.widgets[key].setText('<multiple values>')
                return
            value = new_value
        if value:
            self.widgets[key].setText(value)
        else:
            self.widgets[key].clear()

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        if not selection:
            for key in self.widgets:
                self.widgets[key].clear()
                self.widgets[key].setEnabled(False)
            return
        for key in self.widgets:
            self.widgets[key].setEnabled(True)
            self._update_widget(key)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-13  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import logging
import sys

try:
    import pgi
    pgi.install_as_gi()
except ImportError:
    pass
from gi.repository import GObject, GExiv2

# pydoc gi.repository.GExiv2.Metadata is useful to see methods available

class MetadataHandler(object):
    def __init__(self, path):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.path = path
        self._md = GExiv2.Metadata()
        self._md.open_path(self.path)
        # adopt some GExiv2.Metadata methods
        self.get_exif_tags    = self._md.get_exif_tags
        self.get_iptc_tags    = self._md.get_iptc_tags
        self.get_xmp_tags     = self._md.get_xmp_tags
        self.get_tag_string   = self._md.get_tag_string
        self.get_tag_multiple = self._md.get_tag_multiple
        self.set_tag_string   = self._md.set_tag_string
        self.set_tag_multiple = self._md.set_tag_multiple
        self.clear_tag        = self._md.clear_tag

    def save(self):
        try:
            self._md.save_file(self.path)
        except GObject.GError as ex:
            self.logger.exception(ex)
            return False
        return True

    def copy(self, other, exif=True, iptc=True, xmp=True, comment=True):
        # copy from other to self
        if exif:
            for tag in other._md.get_exif_tags():
                self._md.set_tag_string(
                    tag, other._md.get_tag_string(tag))
        if iptc:
            for tag in other._md.get_iptc_tags():
                self._md.set_tag_multiple(
                    tag, other._md.get_tag_multiple(tag))
        if xmp:
            for tag in other._md.get_xmp_tags():
                self._md.set_tag_multiple(
                    tag, other._md.get_tag_multiple(tag))
        if comment:
            self._md.set_comment(other._md.get_comment())

    def get_tags(self):
        return self.get_exif_tags() + self.get_iptc_tags() + self.get_xmp_tags()



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-13  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import webbrowser

from PyQt4 import QtGui, QtCore

from .photinimap import PhotiniMap

class BingMap(PhotiniMap):
    def __init__(self, *arg, **kw):
        self.copyright_widget = QtGui.QLabel()
        self.copyright_widget.setWordWrap(True)
        PhotiniMap.__init__(self, *arg, **kw)

    def load_api(self):
        return """
    <script charset="UTF-8" type="text/javascript"
      src="http://ecn.dev.virtualearth.net/mapcontrol/mapcontrol.ashx?v=7.0">
    </script>
    <script type="text/javascript">
      var api_key = "%s";
    </script>
""" % 'ArJEzSPM47yeCE31K9ZgelN2jPG20egbQNC8DGM__Z4r9Y8U-hvj4vyHJSRoAcCQ'

    def show_terms(self):
        # return a widget to display map terms and conditions
        result = QtGui.QFrame()
        layout = QtGui.QVBoxLayout()
        result.setLayout(layout)
        layout.addWidget(self.copyright_widget)
        load_tou = QtGui.QPushButton('Terms of Use')
        load_tou.clicked.connect(self.load_tou)
        layout.addWidget(load_tou)
        return result

    @QtCore.pyqtSlot(unicode)
    def new_copyright(self, text):
        self.copyright_widget.setText(text)

    def load_tou(self):
        webbrowser.open_new(
            'http://www.microsoft.com/maps/assets/docs/terms.aspx')



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import logging
import os
import time
import webbrowser

import flickrapi
from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from .utils import Busy

api2 = flickrapi.__version__ >= '2.0'

class FileWithCallback(object):
    def __init__(self, path, callback):
        self.file = open(path, 'rb')
        self.callback = callback
        # the following attributes and methods are used by
        # requests_tools.MultipartEncoder
        self.len = os.path.getsize(path)
        self.fileno = self.file.fileno
        self.tell = self.file.tell

    def read(self, size):
        if self.callback:
            self.callback(self.tell() * 100 // self.len)
        return self.file.read(size)

class UploadThread(QtCore.QThread):
    def __init__(self, flickr, upload_list, photosets, new_photosets):
        QtCore.QThread.__init__(self)
        self.flickr = flickr
        self.upload_list = upload_list
        self.photosets = photosets
        self.new_photosets = new_photosets

    reload_sets = QtCore.pyqtSignal()
    def run(self):
        logger = logging.getLogger(self.__class__.__name__)
        self.file_count = 0
        for params in self.upload_list:
            if api2:
                params['fileobj'] = FileWithCallback(params['filename'], self.callback)
            else:
                params['callback'] = self.callback
            rsp = self.flickr.upload(**params)
            if rsp.attrib['stat'] == 'ok':
                photo_id = rsp.find('photoid').text
                for photoset_id in self.photosets:
                    self.flickr.photosets_addPhoto(
                        photo_id=photo_id, photoset_id=photoset_id)
                if self.new_photosets:
                    for title, description in self.new_photosets:
                        rsp2 = self.flickr.photosets_create(
                            title=title, description=description,
                            primary_photo_id=photo_id)
                        if rsp2.attrib['stat'] == 'ok':
                            self.photosets.append(
                                rsp2.find('photoset').attrib['id'])
                    self.new_photosets = []
                    self.reload_sets.emit()
            else:
                logger.error(
                    'Upload %s failed: %s',
                    os.path.basename(params['filename']), rsp.attrib['stat'])
            self.file_count += 1

    progress_report = QtCore.pyqtSignal(float, float)
    def callback(self, progress, done=None):
        total_progress = (
            (self.file_count * 100) + progress) / len(self.upload_list)
        self.progress_report.emit(progress, total_progress)

class FlickrUploader(QtGui.QWidget):
    def __init__(self, config_store, image_list, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.image_list = image_list
        self.setLayout(QtGui.QGridLayout())
        self.flickr = None
        self.photosets = list()
        self.uploader = None
        self.new_sets = []
        # privacy settings
        self.privacy = dict()
        privacy_group = QtGui.QGroupBox('Who can see the photos?')
        privacy_group.setLayout(QtGui.QVBoxLayout())
        self.privacy['private'] = QtGui.QRadioButton('Only you')
        privacy_group.layout().addWidget(self.privacy['private'])
        ff_group = QtGui.QGroupBox()
        ff_group.setFlat(True)
        ff_group.setLayout(QtGui.QVBoxLayout())
        self.privacy['friends'] = QtGui.QCheckBox('Your friends')
        ff_group.layout().addWidget(self.privacy['friends'])
        self.privacy['family'] = QtGui.QCheckBox('Your family')
        ff_group.layout().addWidget(self.privacy['family'])
        privacy_group.layout().addWidget(ff_group)
        self.privacy['public'] = QtGui.QRadioButton('Anyone')
        self.privacy['public'].toggled.connect(self.enable_ff)
        self.privacy['public'].setChecked(True)
        privacy_group.layout().addWidget(self.privacy['public'])
        # hidden
        self.hidden = QtGui.QCheckBox('Hidden from search')
        privacy_group.layout().addWidget(self.hidden)
        privacy_group.layout().addStretch(1)
        self.layout().addWidget(privacy_group, 0, 0, 3, 2)
        # content type
        self.content_type = dict()
        content_group = QtGui.QGroupBox('Content type')
        content_group.setLayout(QtGui.QVBoxLayout())
        self.content_type['photo'] = QtGui.QRadioButton('Photo')
        self.content_type['photo'].setChecked(True)
        content_group.layout().addWidget(self.content_type['photo'])
        self.content_type['screenshot'] = QtGui.QRadioButton('Screenshot')
        content_group.layout().addWidget(self.content_type['screenshot'])
        self.content_type['other'] = QtGui.QRadioButton('Art/Illustration')
        content_group.layout().addWidget(self.content_type['other'])
        content_group.layout().addStretch(1)
        self.layout().addWidget(content_group, 0, 2, 2, 1)
        # create new set
        new_set_button = QtGui.QPushButton('New set')
        new_set_button.clicked.connect(self.new_set)
        self.layout().addWidget(new_set_button, 2, 2)
        # list of sets widget
        sets_group = QtGui.QGroupBox('Add to sets')
        sets_group.setLayout(QtGui.QVBoxLayout())
        self.scrollarea = QtGui.QScrollArea()
        self.scrollarea.setFrameStyle(QtGui.QFrame.NoFrame)
        sets_group.layout().addWidget(self.scrollarea)
        self.layout().addWidget(sets_group, 0, 3, 3, 1)
        # 'go' button
        self.upload_button = QtGui.QPushButton('Upload\nnow')
        self.upload_button.setEnabled(False)
        self.upload_button.clicked.connect(self.upload)
        self.layout().addWidget(self.upload_button, 2, 4)
        # progress bars
        self.layout().addWidget(QtGui.QLabel('Progress'), 3, 0)
        self.total_progress = QtGui.QProgressBar()
        self.total_progress.sizePolicy().setHorizontalPolicy(
            QtGui.QSizePolicy.Expanding)
        self.layout().addWidget(self.total_progress, 3, 1, 1, 4)
        # adjust spacing
        self.layout().setColumnStretch(1, 1)
        self.layout().setColumnStretch(3, 100)
        self.layout().setRowStretch(1, 1)

    def refresh(self):
        if self.flickr:
            return
        self.get_photosets()

    @QtCore.pyqtSlot()
    def get_photosets(self):
        if not self.authorise():
            return
        with Busy():
            self.photosets = []
            sets = self.flickr.photosets_getList()
        for item in sets.find('photosets').findall('photoset'):
            self.photosets.append({
                'id'    : item.attrib['id'],
                'title' : item.find('title').text,
                })
        sets_widget = QtGui.QWidget()
        self.sets = QtGui.QVBoxLayout()
        sets_widget.setLayout(self.sets)
        for item in self.photosets:
            self.sets.addWidget(
                QtGui.QCheckBox(item['title'].replace('&', '&&')))
        self.scrollarea.setWidget(sets_widget)
        sets_widget.setAutoFillBackground(False)

    @QtCore.pyqtSlot(bool)
    def enable_ff(self, value):
        self.privacy['friends'].setEnabled(self.privacy['private'].isChecked())
        self.privacy['family'].setEnabled(self.privacy['private'].isChecked())

    @QtCore.pyqtSlot()
    def new_set(self):
        dialog = QtGui.QDialog()
        dialog.setWindowTitle('Create new Flickr set')
        dialog.setLayout(QtGui.QFormLayout())
        title = QtGui.QLineEdit()
        dialog.layout().addRow('Title', title)
        description = QtGui.QPlainTextEdit()
        dialog.layout().addRow('Description', description)
        dialog.layout().addRow(QtGui.QLabel(
            'Set will be created when photos are uploaded'))
        buttons = QtGui.QHBoxLayout()
        buttons.addStretch(3)
        cancel = QtGui.QPushButton('cancel')
        cancel.setAutoDefault(False)
        cancel.clicked.connect(dialog.reject)
        buttons.addWidget(cancel)
        OK = QtGui.QPushButton('OK')
        OK.setAutoDefault(False)
        OK.clicked.connect(dialog.accept)
        buttons.addWidget(OK)
        buttons.setStretchFactor(cancel, 1)
        buttons.setStretchFactor(OK, 1)
        dialog.layout().addRow(buttons)
        if dialog.exec_() != QtGui.QDialog.Accepted:
            return
        title = unicode(title.text())
        if not title:
            return
        description = unicode(description.toPlainText())
        self.new_sets.append((title, description))
        check_box = QtGui.QCheckBox(title.replace('&', '&&'))
        check_box.setChecked(True)
        check_box.setEnabled(False)
        self.sets.insertWidget(0, check_box)

    @QtCore.pyqtSlot()
    def upload(self):
        if not self.image_list.unsaved_files_dialog(with_discard=False):
            return
        is_public = ('0', '1')[self.privacy['public'].isChecked()]
        is_family = ('0', '1')[self.privacy['private'].isChecked() and
                               self.privacy['family'].isChecked()]
        is_friend = ('0', '1')[self.privacy['private'].isChecked() and
                               self.privacy['friends'].isChecked()]
        if self.content_type['photo'].isChecked():
            content_type = '1'
        elif self.content_type['screenshot'].isChecked():
            content_type = '2'
        else:
            content_type = '3'
        hidden = ('1', '2')[self.hidden.isChecked()]
        photosets = list()
        for idx in range(len(self.photosets)):
            if self.sets.itemAt(idx + len(self.new_sets)).widget().isChecked():
                photosets.append(self.photosets[idx]['id'])
        new_photosets = self.new_sets
        self.new_sets = []
        # make list of items to upload
        upload_list = list()
        for image in self.image_list.get_selected_images():
            title = image.metadata.get_item('title').as_str()
            if not title:
                title = os.path.basename(image.path)
            description = image.metadata.get_item('description').as_str()
            tags = image.metadata.get_item('keywords')
            if tags.empty():
                tags = ''
            else:
                tags = ' '.join(map(lambda x: '"%s"' % x, tags.value))
            upload_list.append({
                'filename'     : image.path,
                'title'        : title,
                'description'  : description,
                'tags'         : tags,
                'is_public'    : is_public,
                'is_friend'    : is_friend,
                'is_family'    : is_family,
                'content_type' : content_type,
                'hidden'       : hidden,
                })
        # pass the list to a separate thread, so GUI can continue
        if self.authorise():
            self.upload_button.setEnabled(False)
            self.uploader = UploadThread(
                self.flickr, upload_list, photosets, new_photosets)
            self.uploader.progress_report.connect(self.upload_progress)
            self.uploader.finished.connect(self.upload_done)
            self.uploader.reload_sets.connect(self.get_photosets)
            self.uploader.start()
            # we've passed the flickr API object to a new thread, so
            # create a new one for safety
            self.flickr = None
            self.authorise()

    @QtCore.pyqtSlot(float, float)
    def upload_progress(self, progress, total_progress):
        self.total_progress.setValue(total_progress)

    @QtCore.pyqtSlot()
    def upload_done(self):
        self.upload_button.setEnabled(True)
        self.total_progress.setValue(0)
        self.uploader = None

    def authorise(self):
        if self.flickr:
            return True
        api_key = 'b6263c4693e3406aadcfaebe005280a5'
        api_secret = '1e0d912f586d0ed1'
        if api2:
            api_key = unicode(api_key)
            api_secret = unicode(api_secret)
            token        = self.config_store.get('flickr', 'token', '')
            token_secret = self.config_store.get('flickr', 'token_secret', '')
            token = flickrapi.auth.FlickrAccessToken(
                token, token_secret, 'write')
        else:
            token = self.config_store.get('flickr', 'token', '')
        self.flickr = flickrapi.FlickrAPI(
            api_key, api_secret, token=token, store_token=False)
        if api2:
            if self.flickr.token_valid(perms='write'):
                return True
        else:
            token, frob = self.flickr.get_token_part_one(perms='write')
            if token:
                token = self.flickr.get_token_part_two((token, frob))
            if token:
                return True
        if api2:
            self.flickr.get_request_token(oauth_callback='oob')
            auth_url = self.flickr.auth_url(perms='write')
            if webbrowser.open(auth_url, new=2, autoraise=0):
                info_text = 'use your web browser'
            else:
                info_text = 'open "%s" in a web browser' % auth_url
            auth_code, OK = QtGui.QInputDialog.getText(
                self,
                'Photini: authorise Flickr',
                'Please %s to grant access to Photini,\n' % (info_text) +
                'then enter the verification code:')
            if not OK:
                self.flickr = None
                return False
            try:
                self.flickr.get_access_token(unicode(auth_code))
            except Exception, ex:
                self.flickr = None
                return False
        else:
            button = QtGui.QMessageBox.question(
                self,
                'Photini: authorise Flickr',
                'Please use your web browser to grant\n' +
                'access to Photini, then click OK.')
            try:
                token = self.flickr.get_token_part_two((token, frob))
            except flickrapi.exceptions.FlickrError:
                self.flickr = None
                return False
        if api2:
            token = self.flickr.token_cache.token
            self.config_store.set('flickr', 'token',        token.token)
            self.config_store.set('flickr', 'token_secret', token.token_secret)
        else:
            self.config_store.set('flickr', 'token', token)
        return True

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        self.upload_button.setEnabled(
            len(selection) > 0 and self.flickr and not self.uploader)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from datetime import datetime
import fractions
import locale
import logging
import os
import sys

from PyQt4 import QtCore

try:
    from .metadata_gexiv2 import MetadataHandler
except ImportError as e:
    try:
        from .metadata_pyexiv2 import MetadataHandler
    except ImportError:
        # raise exception on the one we really wanted
        raise e
from . import __version__

_encodings = None
def _decode_string(value):
    global _encodings
    if sys.version_info[0] >= 3:
        return value
    if not _encodings:
        _encodings = ['utf_8', 'latin_1']
        char_set = locale.getdefaultlocale()[1]
        if char_set:
            _encodings.append(char_set)
    for encoding in _encodings:
        try:
            return value.decode(encoding)
        except UnicodeDecodeError:
            continue
    return value.decode('utf_8')

def _encode_string(value, max_bytes=None):
    result = value.encode('utf_8')
    if max_bytes:
        result = result[:max_bytes]
    if sys.version_info[0] >= 3:
        result = result.decode('utf_8')
    return result

class NullValue(object):
    def empty(self):
        return True

    def as_str(self):
        return u''

class BaseValue(object):
    def __init__(self, tag=None):
        self.tag = tag
        self.value = None

    def empty(self):
        return self.value is None

    def as_str(self):
        return str(self.value)

    def merge(self, other):
        return self

    def set_value(self, value):
        self.value = value

    def iptc_pred(self, max_bytes):
        return self.as_str()

class IntValue(BaseValue):
    # value is a single integer
    def to_exif(self, md, tag):
        if self.value is None:
            md.clear_tag(tag)
            return
        md.set_tag_string(tag, self.as_str())

    def from_exif(self, md):
        if self.tag not in md.get_exif_tags():
            return
        self.value = int(md.get_tag_string(self.tag))

    def from_xmp(self, md):
        if self.tag not in md.get_xmp_tags():
            return
        self.value = int(md.get_tag_multiple(self.tag)[0])

class LatLongValue(BaseValue):
    # value is a latitude + longitude pair, stored as floating point numbers
    def as_str(self):
        if self.value:
            return '%.6f, %.6f' % self.value
        return u''

    def to_exif(self, md, tag):
        lat_tag = tag
        long_tag = lat_tag.replace('Latitude', 'Longitude')
        if self.value is None:
            md.clear_tag(lat_tag)
            md.clear_tag(lat_tag + 'Ref')
            md.clear_tag(long_tag)
            md.clear_tag(long_tag + 'Ref')
            return
        for this_tag, value, sign_char in zip(
                            (lat_tag, long_tag), self.value, ('NS', 'EW')):
            if value >= 0.0:
                ref_string = sign_char[0]
            else:
                ref_string = sign_char[1]
                value = -value
            degrees = int(value)
            value = (value - degrees) * 60.0
            minutes = int(value)
            seconds = (value - minutes) * 60.0
            seconds = fractions.Fraction(seconds).limit_denominator(1000000)
            value_string = '%d/1 %d/1 %d/%d' % (
                degrees, minutes, seconds.numerator, seconds.denominator)
            md.set_tag_string(this_tag, value_string)
            md.set_tag_string(this_tag + 'Ref', ref_string)

    def from_exif(self, md):
        lat_tag = self.tag
        long_tag = lat_tag.replace('Latitude', 'Longitude')
        result = []
        for tag in (lat_tag, long_tag):
            ref_tag = tag + 'Ref'
            if not (tag in md.get_exif_tags() and ref_tag in md.get_exif_tags()):
                return
            value_string = md.get_tag_string(tag)
            ref_string = md.get_tag_string(ref_tag)
            parts = map(fractions.Fraction, value_string.split())
            value = float(parts[0])
            if len(parts) > 1:
                value += float(parts[1]) / 60.0
            if len(parts) > 2:
                value += float(parts[2]) / 3600.0
            if ref_string in ('S', 'W'):
                value = -value
            result.append(value)
        self.value = (result[0], result[1])

    def from_xmp(self, md):
        lat_tag = self.tag
        long_tag = lat_tag.replace('Latitude', 'Longitude')
        result = []
        for tag in (lat_tag, long_tag):
            if not tag in md.get_xmp_tags():
                return
            value_string = md.get_tag_multiple(tag)[0]
            degrees, residue = value_string.split(',')
            minutes = residue[:-1]
            ref_string = residue[-1]
            value = float(degrees) + (float(minutes) / 60.0)
            if ref_string in ('S', 'W'):
                value = -value
            result.append(value)
        self.value = (result[0], result[1])

class StringValue(BaseValue):
    # value is a single unicode string
    def __init__(self, tag=None):
        BaseValue.__init__(self, tag)
        self.value = u''

    def empty(self):
        return len(self.value) == 0

    def sanitise(self):
        self.value = self.value.strip()

    def iptc_pred(self, max_bytes):
        return _decode_string(_encode_string(self.value, max_bytes))

    def as_str(self):
        return self.value

    def merge(self, other):
        if isinstance(other, ListValue):
            return other.merge(self)
        result = StringValue()
        result.value = self.value
        if other.value not in result.value:
            result.value = '%s // %s' % (result.value, other.value)
        return result

    def set_value(self, value):
        self.value = value
        self.sanitise()

    def to_exif(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        md.set_tag_string(tag, _encode_string(self.value))

    def from_exif(self, md):
        if self.tag not in md.get_exif_tags():
            return
        self.value = _decode_string(md.get_tag_string(self.tag))
        self.sanitise()

    def to_iptc(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        md.set_tag_multiple(tag, [_encode_string(self.value)])

    def from_iptc(self, md):
        if self.tag not in md.get_iptc_tags():
            return
        self.value = md.get_tag_multiple(self.tag)
        self.value = _decode_string(self.value[0])
        self.sanitise()

    def to_xmp(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        md.set_tag_multiple(tag, [self.value])

    def from_xmp(self, md):
        if self.tag not in md.get_xmp_tags():
            return
        self.value = md.get_tag_multiple(self.tag)[0]
        if sys.version_info[0] < 3 and not isinstance(self.value, unicode):
            self.value = self.value.decode('utf_8')
        self.sanitise()

class ListValue(BaseValue):
    # value is an array of unicode strings
    def __init__(self, tag=None):
        BaseValue.__init__(self, tag)
        self.value = []

    def empty(self):
        return len(self.value) == 0

    def sanitise(self):
        new_value = []
        for item in self.value:
            item = item.strip()
            if item:
                new_value.append(item)
        self.value = new_value

    def iptc_pred(self, max_bytes):
        pred_value = map(
            lambda x: _decode_string(_encode_string(x, max_bytes)),
            self.value)
        return u'; '.join(pred_value)

    def as_str(self):
        return u'; '.join(self.value)

    def merge(self, other):
        result = ListValue()
        result.value = list(self.value)
        if isinstance(other, StringValue):
            items = [other.value]
        else:
            items = list(other.value)
        for item in items:
            if item not in result.value:
                result.value.append(item)
        return result

    def set_value(self, value):
        if value:
            self.value = value.split(';')
            self.sanitise()
        else:
            self.value = []

    def to_exif(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        md.set_tag_string(tag, _encode_string(self.as_str()))

    def to_iptc(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        md.set_tag_multiple(tag, map(_encode_string, self.value))

    def from_iptc(self, md):
        if self.tag not in md.get_exif_tags():
            return
        self.value = map(_decode_string, md.get_tag_multiple(self.tag))
        self.sanitise()

    def to_xmp(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        md.set_tag_multiple(tag, self.value)

    def from_xmp(self, md):
        if self.tag not in md.get_xmp_tags():
            return
        self.value = md.get_tag_multiple(self.tag)
        if sys.version_info[0] < 3 and self.value and not isinstance(self.value[0], unicode):
            self.value = map(lambda x: x.decode('utf_8'), self.value)
        self.sanitise()

class DateTimeValue(BaseValue):
    # value is a Python datetime object
    def as_str(self):
        if self.value:
            return self.value.isoformat()
        return u''

    def to_exif(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        string_value = self.value.strftime('%Y:%m:%d %H:%M:%S')
        md.set_tag_string(tag, string_value)

    def from_exif(self, md):
        if self.tag not in md.get_exif_tags():
            return
        string_value = md.get_tag_string(self.tag)
        self.value = datetime.strptime(string_value, '%Y:%m:%d %H:%M:%S')

    def to_iptc(self, md, tag):
        date_tag = tag
        time_tag = date_tag.replace('Date', 'Time')
        if not self.value:
            md.clear_tag(date_tag)
            md.clear_tag(time_tag)
            return
        date_string = self.value.strftime('%Y-%m-%d')
        time_string = self.value.strftime('%H:%M:%S')
        md.set_tag_multiple(date_tag, [date_string])
        md.set_tag_multiple(time_tag, [time_string])

    def from_iptc(self, md):
        date_tag = self.tag
        time_tag = date_tag.replace('Date', 'Time')
        if date_tag in md.get_iptc_tags():
            date_string = md.get_tag_multiple(date_tag)[0]
            has_date = True
        else:
            date_string = '0001-01-01'
            has_date = False
        if time_tag in md.get_iptc_tags():
            time_string = md.get_tag_multiple(time_tag)[0][:8]
            has_time = True
        else:
            time_string = '00:00:00'
            has_time = False
        if has_date or has_time:
            self.value = datetime.strptime(
                date_string + time_string, '%Y-%m-%d%H:%M:%S')

    def to_xmp(self, md, tag):
        if not self.value:
            md.clear_tag(tag)
            return
        string_value = self.value.strftime('%Y-%m-%dT%H:%M:%S')
        md.set_tag_multiple(tag, [string_value])

    def from_xmp(self, md):
        if self.tag not in md.get_xmp_tags():
            return
        string_value = md.get_tag_multiple(self.tag)[0]
        # remove any time zone info
        if string_value.count(':') >= 2:
            string_value = string_value[:19]
        else:
            string_value = string_value[:16]
        # extend short strings to include all info
        string_value += '0001-01-01T00:00:00'[len(string_value):]
        # convert to datetime
        self.value = datetime.strptime(string_value, '%Y-%m-%dT%H:%M:%S')

# class to use for each tag's data
_data_object = {
    'Exif.GPSInfo.GPSLatitude'           : LatLongValue,
    'Exif.Image.Artist'                  : StringValue,
    'Exif.Image.Copyright'               : StringValue,
    'Exif.Image.DateTime'                : DateTimeValue,
    'Exif.Image.DateTimeOriginal'        : DateTimeValue,
    'Exif.Image.ImageDescription'        : StringValue,
    'Exif.Image.Orientation'             : IntValue,
    'Exif.Image.ProcessingSoftware'      : StringValue,
    'Exif.Photo.DateTimeDigitized'       : DateTimeValue,
    'Exif.Photo.DateTimeOriginal'        : DateTimeValue,
    'Iptc.Application2.Byline'           : ListValue,
    'Iptc.Application2.Caption'          : StringValue,
    'Iptc.Application2.Copyright'        : StringValue,
    'Iptc.Application2.DateCreated'      : DateTimeValue,
    'Iptc.Application2.DigitizationDate' : DateTimeValue,
    'Iptc.Application2.Headline'         : StringValue,
    'Iptc.Application2.Keywords'         : ListValue,
    'Iptc.Application2.ObjectName'       : StringValue,
    'Xmp.dc.creator'                     : ListValue,
    'Xmp.dc.description'                 : StringValue,
    'Xmp.dc.rights'                      : StringValue,
    'Xmp.dc.subject'                     : ListValue,
    'Xmp.dc.title'                       : StringValue,
    'Xmp.photoshop.DateCreated'          : DateTimeValue,
    'Xmp.exif.DateTimeDigitized'         : DateTimeValue,
    'Xmp.exif.DateTimeOriginal'          : DateTimeValue,
    'Xmp.exif.GPSLatitude'               : LatLongValue,
    'Xmp.tiff.Artist'                    : StringValue,
    'Xmp.tiff.Copyright'                 : StringValue,
    'Xmp.tiff.DateTime'                  : DateTimeValue,
    'Xmp.tiff.ImageDescription'          : StringValue,
    'Xmp.tiff.Orientation'               : IntValue,
    'Xmp.xmp.CreateDate'                 : DateTimeValue,
    'Xmp.xmp.ModifyDate'                 : DateTimeValue,
    }
# some data requires more than one tag
_associated_tags = {
    'Exif.GPSInfo.GPSLatitude'           : ('Exif.GPSInfo.GPSLatitudeRef',
                                            'Exif.GPSInfo.GPSLongitude',
                                            'Exif.GPSInfo.GPSLongitudeRef'),
    'Iptc.Application2.DateCreated'      : ('Iptc.Application2.TimeCreated',),
    'Iptc.Application2.DigitizationDate' : ('Iptc.Application2.DigitizationTime',),
    'Xmp.exif.GPSLatitude'               : ('Xmp.exif.GPSLongitude',),
    }
# maximum length of Iptc data
_max_bytes = {
    'Iptc.Application2.Byline'           :   32,
    'Iptc.Application2.Caption'          : 2000,
    'Iptc.Application2.Copyright'        :  128,
    'Iptc.Application2.DateCreated'      : None,
    'Iptc.Application2.DigitizationDate' : None,
    'Iptc.Application2.Headline'         :  256,
    'Iptc.Application2.Keywords'         :   64,
    'Iptc.Application2.ObjectName'       :   64,
    }

class Metadata(QtCore.QObject):
    # mapping of preferred tags to Photini data fields
    _primary_tags = {
        'copyright'      : {'Exif' : 'Exif.Image.Copyright',
                            'Xmp'  : 'Xmp.dc.rights',
                            'Iptc' : 'Iptc.Application2.Copyright'},
        'creator'        : {'Exif' : 'Exif.Image.Artist',
                            'Xmp'  : 'Xmp.dc.creator',
                            'Iptc' : 'Iptc.Application2.Byline'},
        'date_digitised' : {'Exif' : 'Exif.Photo.DateTimeDigitized',
                            'Xmp'  : 'Xmp.xmp.CreateDate',
                            'Iptc' : 'Iptc.Application2.DigitizationDate'},
        'date_modified'  : {'Exif' : 'Exif.Image.DateTime',
                            'Xmp'  : 'Xmp.xmp.ModifyDate'},
        'date_taken'     : {'Exif' : 'Exif.Photo.DateTimeOriginal',
                            'Xmp'  : 'Xmp.photoshop.DateCreated',
                            'Iptc' : 'Iptc.Application2.DateCreated'},
        'description'    : {'Exif' : 'Exif.Image.ImageDescription',
                            'Xmp'  : 'Xmp.dc.description',
                            'Iptc' : 'Iptc.Application2.Caption'},
        'keywords'       : {'Xmp'  : 'Xmp.dc.subject',
                            'Iptc' : 'Iptc.Application2.Keywords'},
        'latlong'        : {'Exif' : 'Exif.GPSInfo.GPSLatitude'},
        'orientation'    : {'Exif' : 'Exif.Image.Orientation'},
        'software'       : {'Exif' : 'Exif.Image.ProcessingSoftware'},
        'title'          : {'Xmp'  : 'Xmp.dc.title',
                            'Iptc' : 'Iptc.Application2.ObjectName'},
        }
    # mapping of duplicate tags to Photini data fields
    # data in these is merged in when data is read
    # they get deleted when data is written
    _secondary_tags = {
        'copyright'      : {'Xmp'  : ('Xmp.tiff.Copyright',)},
        'creator'        : {'Xmp'  : ('Xmp.tiff.Artist',)},
        'date_digitised' : {'Xmp'  : ('Xmp.exif.DateTimeDigitized',)},
        'date_modified'  : {'Xmp'  : ('Xmp.tiff.DateTime',)},
        'date_taken'     : {'Exif' : ('Exif.Image.DateTimeOriginal',),
                            'Xmp'  : ('Xmp.exif.DateTimeOriginal',)},
        'description'    : {'Xmp'  : ('Xmp.tiff.ImageDescription',)},
        'keywords'       : {},
        'latlong'        : {'Xmp'  : ('Xmp.exif.GPSLatitude',)},
        'orientation'    : {'Xmp'  : ('Xmp.tiff.Orientation',)},
        'software'       : {},
        'title'          : {'Iptc' : ('Iptc.Application2.Headline',)},
        }
    def __init__(self, path, parent=None):
        QtCore.QObject.__init__(self, parent)
        self.logger = logging.getLogger(self.__class__.__name__)
        # create metadata handlers for image file and sidecar (if present)
        self._path = path
        self._if = MetadataHandler(path)
        self._sc_path = self._find_side_car(path)
        if self._sc_path:
            self._sc = MetadataHandler(self._sc_path)
        else:
            self._sc = None
        self._unsaved = False
        self._value_cache = {}

    def _find_side_car(self, path):
        for base in (os.path.splitext(path)[0], path):
            for ext in ('.xmp', '.XMP'):
                result = base + ext
                if os.path.exists(result):
                    return result
        return None

    def create_side_car(self):
        self._sc_path = self._path + '.xmp'
        with open(self._sc_path, 'w') as of:
            of.write('<x:xmpmeta x:xmptk="XMP Core 4.4.0-Exiv2" ')
            of.write('xmlns:x="adobe:ns:meta/">\n')
            of.write('</x:xmpmeta>')
        self._sc = MetadataHandler(self._sc_path)
        self._sc.copy(self._if, comment=False)

    def save(self, if_mode, sc_mode):
        if not self._unsaved:
            return
        self.set_item('software', 'Photini editor v%s' % (__version__))
        if sc_mode == 'delete' and self._sc:
            self._if.copy(self._sc, comment=False)
        OK = False
        if if_mode:
            OK = self._if.save()
        if sc_mode == 'delete' and self._sc and OK:
            os.unlink(self._sc_path)
            self._sc = None
        if sc_mode == 'auto' and not self._sc and not OK:
            self.create_side_car()
        if sc_mode == 'always' and not self._sc:
            self.create_side_car()
        if self._sc:
            OK = self._sc.save()
        self._set_unsaved(not OK)

    # tag lists: merge tags from sidecar and image file
    def get_exif_tags(self):
        result = self._if.get_exif_tags()
        if self._sc:
            for tag in self._sc.get_exif_tags():
                if tag not in result:
                    result.append(tag)
        return result

    def get_iptc_tags(self):
        result = self._if.get_iptc_tags()
        if self._sc:
            for tag in self._sc.get_iptc_tags():
                if tag not in result:
                    result.append(tag)
        return result

    def get_xmp_tags(self):
        result = self._if.get_xmp_tags()
        if self._sc:
            for tag in self._sc.get_xmp_tags():
                if tag not in result:
                    result.append(tag)
        return result

    # getters: use sidecar if tag is present, otherwise use image file
    def get_tag_string(self, tag):
        if self._sc and tag in self._sc.get_tags():
            return self._sc.get_tag_string(tag)
        return self._if.get_tag_string(tag)

    def get_tag_multiple(self, tag):
        if self._sc and tag in self._sc.get_tags():
            return self._sc.get_tag_multiple(tag)
        return self._if.get_tag_multiple(tag)

    # setters: set in both sidecar and image file
    def set_tag_string(self, tag, value):
        if self._sc:
            self._sc.set_tag_string(tag, value)
        self._if.set_tag_string(tag, value)

    def set_tag_multiple(self, tag, value):
        if self._sc:
            self._sc.set_tag_multiple(tag, value)
        self._if.set_tag_multiple(tag, value)

    def clear_tag(self, tag):
        if self._sc:
            self._sc.clear_tag(tag)
        self._if.clear_tag(tag)

    def get_tags(self):
        return self.get_exif_tags() + self.get_iptc_tags() + self.get_xmp_tags()

    def _get_value(self, family, tag):
        result = _data_object[tag](tag)
        if family == 'Exif':
            result.from_exif(self)
        elif family == 'Iptc':
            result.from_iptc(self)
        else:
            result.from_xmp(self)
        return result

    def get_item(self, name):
        if name in self._value_cache:
            return self._value_cache[name]
        # get values from all 3 families, using first tag in list that has data
        value = {'Exif': NullValue(), 'Iptc': NullValue(), 'Xmp': NullValue()}
        for family in self._primary_tags[name]:
            try:
                value[family] = self._get_value(
                    family, self._primary_tags[name][family])
            except Exception, ex:
                self.logger.exception(ex)
        for family in self._secondary_tags[name]:
            for tag in self._secondary_tags[name][family]:
                try:
                    new_value = self._get_value(family, tag)
                except Exception, ex:
                    self.logger.exception(ex)
                    continue
                if new_value.empty():
                    continue
                elif value[family].empty():
                    value[family] = new_value
                elif new_value.as_str() not in value[family].as_str():
                    self.logger.warning('merging %s with %s',
                                        new_value.tag, value[family].tag)
                    value[family] = value[family].merge(new_value)
        # choose preferred family
        if not value['Exif'].empty():
            preference = 'Exif'
        elif not value['Xmp'].empty():
            preference = 'Xmp'
        else:
            preference = 'Iptc'
        # check for IPTC being modified by non-compliant software
        if preference != 'Iptc' and not value['Iptc'].empty():
            iptc_tag = value['Iptc'].tag
            pred_str = value[preference].iptc_pred(_max_bytes[iptc_tag])
            if pred_str != value['Iptc'].as_str():
                self.logger.warning(
                    '%s mismatch with %s', iptc_tag, value[preference].tag)
                preference = 'Iptc'
        # merge in non-matching data so user can review it
        result = value[preference]
        for family in ('Exif', 'Xmp'):
            if preference != family and not value[family].empty():
                if value[family].as_str() not in result.as_str():
                    self.logger.warning('merging %s with %s',
                                        value[family].tag, value[preference].tag)
                    result = result.merge(value[family])
        self._value_cache[name] = result
        return result

    def set_item(self, name, value):
        current_object = self.get_item(name)
        for family in ('Exif', 'Xmp', 'Iptc'):
            if family in self._primary_tags[name]:
                tag = self._primary_tags[name][family]
                new_object = _data_object[tag](tag)
                break
        new_object.set_value(value)
        if new_object.as_str() == current_object.as_str():
            return
        self._value_cache[name] = new_object
        # write data to primary tags (iptc only if it already exists)
        for family in self._primary_tags[name]:
            tag = self._primary_tags[name][family]
            if family == 'Exif':
                new_object.to_exif(self, tag)
            elif family == 'Xmp':
                new_object.to_xmp(self, tag)
            elif tag in self.get_iptc_tags():
                new_object.to_iptc(self, tag)
        # delete secondary tags
        for family in self._secondary_tags[name]:
            for tag in self._secondary_tags[name][family]:
                if tag in self.get_tags():
                    self.clear_tag(tag)
                if tag in _associated_tags:
                    for sup_tag in _associated_tags[tag]:
                        if sup_tag in self.get_tags():
                            self.clear_tag(sup_tag)
        self._set_unsaved(True)

    def del_item(self, name):
        self.set_item(name, None)

    new_status = QtCore.pyqtSignal(bool)
    def _set_unsaved(self, status):
        self._unsaved = status
        self.new_status.emit(self._unsaved)

    def changed(self):
        return self._unsaved



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import logging
import os
import webbrowser

from PyQt4 import QtGui, QtCore, QtWebKit
from PyQt4.QtCore import Qt

from .utils import data_dir
from . import __version__

class WebPage(QtWebKit.QWebPage):
    def __init__(self, parent=None):
        QtWebKit.QWebPage.__init__(self, parent)
        self.logger = logging.getLogger(self.__class__.__name__)

    def javaScriptConsoleMessage(self, msg, line, source):
        if unicode(msg).startswith(
                        "Consider using 'dppx' units instead of 'dpi'"):
            return
        self.logger.error('%s line %d: %s', source, line, msg)

    def userAgentForUrl(self, url):
        # Nominatim requires the user agent to identify the application
        if url.host() == 'nominatim.openstreetmap.org':
            return 'Photini/%s' % __version__
        return QtWebKit.QWebPage.userAgentForUrl(self, url)

class WebView(QtWebKit.QWebView):
    def dragMoveEvent(self, event):
        if event.mimeData().hasText():
            event.acceptProposedAction()

    drop_text = QtCore.pyqtSignal(int, int, str)
    def dropEvent(self, event):
        if not event.mimeData().hasText():
            return
        text = event.mimeData().text()
        if not str(text).strip():
            return
        event.acceptProposedAction()
        self.drop_text.emit(event.pos().x(), event.pos().y(), text)

class PhotiniMap(QtGui.QWidget):
    def __init__(self, config_store, image_list, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.image_list = image_list
        self.location = dict()
        self.search_string = None
        self.map_loaded = False
        layout = QtGui.QGridLayout()
        layout.setMargin(0)
        layout.setRowStretch(6, 1)
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)
        # map
        self.map = WebView()
        self.map.setPage(WebPage())
        self.map.setAcceptDrops(False)
        self.map.page().setLinkDelegationPolicy(QtWebKit.QWebPage.DelegateAllLinks)
        self.map.page().linkClicked.connect(self.link_clicked)
        self.map.page().loadFinished.connect(self.load_finished)
        self.map.page().mainFrame().addToJavaScriptWindowObject("python", self)
        self.map.drop_text.connect(self.drop_text)
        self.layout().addWidget(self.map, 0, 1, 8, 1)
        # search
        self.layout().addWidget(QtGui.QLabel('Search:'), 0, 0)
        self.edit_box = QtGui.QComboBox()
        self.edit_box.setMinimumWidth(200)
        self.edit_box.setEditable(True)
        self.edit_box.setInsertPolicy(QtGui.QComboBox.NoInsert)
        self.edit_box.lineEdit().setPlaceholderText('<new search>')
        self.edit_box.lineEdit().returnPressed.connect(self.search)
        self.edit_box.activated.connect(self.goto_search_result)
        self.clear_search()
        self.edit_box.setEnabled(False)
        self.layout().addWidget(self.edit_box, 1, 0)
        # latitude & longitude
        self.layout().addWidget(QtGui.QLabel('Latitude, longitude:'), 2, 0)
        self.coords = QtGui.QLineEdit()
        self.coords.editingFinished.connect(self.new_coords)
        self.coords.setEnabled(False)
        self.layout().addWidget(self.coords, 3, 0)
        # load map button
        self.load_map = QtGui.QPushButton('\nLoad map\n')
        self.load_map.clicked.connect(self.initialise)
        self.layout().addWidget(self.load_map, 7, 0)
        # other init
        self.image_list.image_list_changed.connect(self.image_list_changed)

    def link_clicked(self, url):
        webbrowser.open_new(url.toString())

    @QtCore.pyqtSlot()
    def image_list_changed(self):
        self.new_images()
        self.see_all()

    def initialise(self):
        page_start = """
<!DOCTYPE html>
<html>
  <head>
    <meta name="viewport" http-equiv="Content-Type"
      content="initial-scale=1.0, user-scalable=no" />
    <style type="text/css">
      html { height: 100% }
      body { height: 100%; margin: 0; padding: 0 }
      #mapDiv { height: 100% }
    </style>
"""
        page_end = """
    <script type="text/javascript" src="%s.js">
    </script>
  </head>
  <body onload="initialize(%f, %f, %d)">
    <div id="mapDiv" style="width:100%%; height:100%%"></div>
  </body>
</html>
"""
        lat, lng = eval(
            self.config_store.get('map', 'centre', '(51.0, 0.0)'))
        zoom = eval(self.config_store.get('map', 'zoom', '11'))
        self.map.setHtml(
            page_start + self.load_api() +
            page_end % (self.__class__.__name__.lower(), lat, lng, zoom),
            QtCore.QUrl.fromLocalFile(data_dir))

    @QtCore.pyqtSlot(bool)
    def load_finished(self, success):
        if success:
            self.map_loaded = True
            self.layout().removeWidget(self.load_map)
            self.load_map.setParent(None)
            show_terms = self.show_terms()
            show_terms.setStyleSheet('QPushButton, QLabel { font-size: 10px }')
            self.layout().addWidget(show_terms, 7, 0)
            self.edit_box.setEnabled(True)
            self.map.setAcceptDrops(True)
            self.new_images()

    def refresh(self):
        if not self.map_loaded:
            return
        lat, lng = eval(self.config_store.get('map', 'centre'))
        zoom = eval(self.config_store.get('map', 'zoom'))
        self.JavaScript('setView(%s, %s, %d)' % (repr(lat), repr(lng), zoom))
        self.new_images()

    @QtCore.pyqtSlot(float, float, int)
    def new_bounds(self, centre_lat, centre_lng, zoom):
        self.map_centre = centre_lat, centre_lng
        self.config_store.set('map', 'centre', str(self.map_centre))
        self.config_store.set('map', 'zoom', str(zoom))

    @QtCore.pyqtSlot(int, int, str)
    def drop_text(self, x, y, text):
        lat, lng = self.JavaScript(
            'latLngFromPixel(%d, %d)' % (x, y)).toPyObject()
        for path in eval(str(text)):
            image = self.image_list.get_image(path)
            self._add_marker(image, lat, lng)
            self._set_metadata(image, lat, lng)
        self.display_coords()
        self.see_selection()

    def new_coords(self):
        text = str(self.coords.text()).strip()
        if not text:
            for image in self.image_list.get_selected_images():
                image.metadata.del_item('latlong')
            self.JavaScript('delMarker("%s")' % (image.path))
            return
        lat, lng = map(float, text.split(','))
        for image in self.image_list.get_selected_images():
            self._set_metadata(image, lat, lng)
            self._add_marker(image, lat, lng)
        self.display_coords()
        self.see_selection()

    def see_all(self):
        marker_list = list()
        for image in self.image_list.get_images():
            marker_list.append(image.path)
        if not marker_list:
            return
        marker_list = '","'.join(marker_list)
        self.JavaScript('seeMarkers(["%s"])' % (marker_list))

    def see_selection(self):
        marker_list = list()
        for image in self.image_list.get_selected_images():
            marker_list.append(image.path)
        if not marker_list:
            return
        marker_list = '","'.join(marker_list)
        self.JavaScript('seeMarkers(["%s"])' % (marker_list))

    def display_coords(self):
        coords = None
        for image in self.image_list.get_selected_images():
            latlong = image.metadata.get_item('latlong')
            if latlong.empty():
                new_coords = None
            else:
                new_coords = latlong.value
            if coords and new_coords != coords:
                self.coords.setText("<multiple values>")
                return
            coords = new_coords
        if coords:
            self.coords.setText('%.6f, %.6f' % coords)
        else:
            self.coords.clear()

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        if selection:
            self.coords.setEnabled(True)
        else:
            self.coords.setEnabled(False)
        for image in self.image_list.get_images():
            latlong = image.metadata.get_item('latlong')
            if not latlong.empty():
                self.JavaScript(
                    'enableMarker("%s", %d)' % (image.path, image.selected))
        self.display_coords()
        self.see_selection()

    def new_images(self):
        self.JavaScript('removeMarkers()')
        for image in self.image_list.get_images():
            latlong = image.metadata.get_item('latlong')
            if not latlong.empty():
                self._add_marker(image, *latlong.value)

    def _add_marker(self, image, lat, lng):
        if not self.map_loaded:
            return
        if lat is None or lng is None:
            return
        self.JavaScript('addMarker("%s", %s, %s, "%s", %d)' % (
            image.path, repr(lat), repr(lng), image.name, image.selected))

    def search(self, search_string=None):
        if not search_string:
            search_string = self.edit_box.lineEdit().text()
            self.edit_box.clearEditText()
        if not search_string:
            return
        self.search_string = search_string
        self.clear_search()
        self.location = dict()
        self.JavaScript('search("%s")' % (search_string))

    def clear_search(self):
        self.edit_box.clear()
        self.edit_box.addItem('')
        if self.search_string:
            self.edit_box.addItem('<repeat search>')

    @QtCore.pyqtSlot(float, float, unicode)
    def search_result(self, lat, lng, name):
        self.edit_box.addItem(name)
        self.location[unicode(name)] = lat, lng
        self.edit_box.showPopup()

    @QtCore.pyqtSlot(int)
    def goto_search_result(self, idx):
        self.edit_box.setCurrentIndex(0)
        self.edit_box.clearFocus()
        if idx == 0:
            return
        if self.search_string and idx == 1:
            # repeat search
            self.search(self.search_string)
            return
        name = unicode(self.edit_box.itemText(idx))
        if name in self.location:
            lat, lng = self.location[name]
            self.JavaScript('goTo(%s, %s)' % (repr(lat), repr(lng)))

    @QtCore.pyqtSlot(str)
    def marker_click(self, path):
        self.image_list.select_image(str(path))

    @QtCore.pyqtSlot(float, float, str)
    def marker_drag_end(self, lat, lng, path):
        image = self.image_list.get_image(str(path))
        self._set_metadata(image, lat, lng)
        self.display_coords()
        self.see_selection()

    def _set_metadata(self, image, lat, lng):
        image.metadata.set_item('latlong', (lat, lng))

    def JavaScript(self, command):
        if self.map_loaded:
            return self.map.page().mainFrame().evaluateJavaScript(command)
        return None



#!/usr/bin/env python

##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-13  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

"""
usage: editor.py [options]
options are:
  -h       | --help        display this help
  -v       | --verbose     increase number of logging messages
  -V       | --version     display version information and exit
"""

from __future__ import unicode_literals

import logging
from optparse import OptionParser
import os
import sys
if sys.version_info[0] >= 3:
    from urllib.request import getproxies
    from urllib.parse import urlparse
else:
    from urllib import getproxies
    from urlparse import urlparse 
import webbrowser

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt
from PyQt4.QtNetwork import QNetworkProxy

from .configstore import ConfigStore
from .bingmap import BingMap
from .descriptive import Descriptive
try:
    from .flickr import FlickrUploader
except ImportError:
    FlickrUploader = None
from .editsettings import EditSettings
from .googlemap import GoogleMap
try:
    from .importer import Importer
except ImportError:
    Importer = None
from .openstreetmap import OpenStreetMap
from .imagelist import ImageList
from .loggerwindow import LoggerWindow
try:
    from .picasa import PicasaUploader
except ImportError:
    PicasaUploader = None
from .technical import Technical
from .utils import data_dir
from . import __version__

class MainWindow(QtGui.QMainWindow):
    def __init__(self, verbose):
        QtGui.QMainWindow.__init__(self)
        self.setWindowTitle("Photini photo metadata editor")
        self.selection = list()
        # logger window
        self.loggerwindow = LoggerWindow(verbose)
        self.logger = logging.getLogger(self.__class__.__name__)
        # config store
        self.config_store = ConfigStore('editor')
        # set network proxy
        proxies = getproxies()
        if 'http' in proxies:
            parsed = urlparse(proxies['http'])
            QNetworkProxy.setApplicationProxy(QNetworkProxy(
                QNetworkProxy.HttpProxy, parsed.hostname, parsed.port))
        # restore size
        size = self.width(), self.height()
        self.resize(*eval(
            self.config_store.get('main_window', 'size', str(size))))
        # image selector
        self.image_list = ImageList(self.config_store)
        self.image_list.selection_changed.connect(self.new_selection)
        self.image_list.new_metadata.connect(self.new_metadata)
        # prepare list of tabs and associated stuff
        self.tab_list = (
            {'name' : '&Descriptive metadata', 'class' : Descriptive},
            {'name' : '&Technical metadata',   'class' : Technical},
            {'name' : 'Map (&Google)',         'class' : GoogleMap},
            {'name' : 'Map (&Bing)',           'class' : BingMap},
            {'name' : 'Map (&OSM)',            'class' : OpenStreetMap},
            {'name' : '&Flickr upload',        'class' : FlickrUploader},
            {'name' : '&Picasa upload',        'class' : PicasaUploader},
            {'name' : '&Import from camera',   'class' : Importer},
            )
        for tab in self.tab_list:
            tab['key'] = tab['name'].replace('&', '').replace(' ', '_')
            tab['key'] = tab['key'].replace('(', '').replace(')', '').lower()
            if tab['class']:
                tab['object'] = tab['class'](self.config_store, self.image_list)
            else:
                tab['object'] = None
        # file menu
        file_menu = self.menuBar().addMenu('File')
        open_action = QtGui.QAction('Open images', self)
        open_action.setShortcuts(['Ctrl+O'])
        open_action.triggered.connect(self.image_list.open_files)
        file_menu.addAction(open_action)
        self.save_action = QtGui.QAction('Save images with new data', self)
        self.save_action.setShortcuts(['Ctrl+S'])
        self.save_action.setEnabled(False)
        self.save_action.triggered.connect(self.image_list.save_files)
        file_menu.addAction(self.save_action)
        self.close_action = QtGui.QAction('Close selected images', self)
        self.close_action.setEnabled(False)
        self.close_action.triggered.connect(self.close_files)
        file_menu.addAction(self.close_action)
        close_all_action = QtGui.QAction('Close all images', self)
        close_all_action.triggered.connect(self.close_all_files)
        file_menu.addAction(close_all_action)
        file_menu.addSeparator()
        quit_action = QtGui.QAction('Quit', self)
        quit_action.setShortcuts(['Ctrl+Q', 'Ctrl+W'])
        quit_action.triggered.connect(
            QtGui.QApplication.instance().closeAllWindows)
        file_menu.addAction(quit_action)
        # options menu
        options_menu = self.menuBar().addMenu('Options')
        settings_action = QtGui.QAction('Settings', self)
        settings_action.triggered.connect(self.edit_settings)
        options_menu.addAction(settings_action)
        options_menu.addSeparator()
        for tab in self.tab_list:
            tab['action'] = QtGui.QAction(tab['name'].replace('&', ''), self)
            tab['action'].setCheckable(True)
            if tab['class']:
                tab['action'].setChecked(
                    eval(self.config_store.get('tabs', tab['key'], 'True')))
            else:
                tab['action'].setEnabled(False)
            tab['action'].triggered.connect(self.add_tabs)
            options_menu.addAction(tab['action'])
        # help menu
        help_menu = self.menuBar().addMenu('Help')
        about_action = QtGui.QAction('About Photini', self)
        about_action.triggered.connect(self.about)
        help_menu.addAction(about_action)
        help_menu.addSeparator()
        help_action = QtGui.QAction('Photini documentation', self)
        help_action.triggered.connect(self.open_docs)
        help_menu.addAction(help_action)
        # main application area
        self.central_widget = QtGui.QSplitter()
        self.central_widget.setOrientation(Qt.Vertical)
        self.central_widget.setChildrenCollapsible(False)
        self.tabs = QtGui.QTabWidget()
        self.tabs.currentChanged.connect(self.new_tab)
        self.add_tabs()
        self.central_widget.addWidget(self.tabs)
        self.central_widget.addWidget(self.image_list)
        size = self.central_widget.sizes()
        self.central_widget.setSizes(eval(
            self.config_store.get('main_window', 'split', str(size))))
        self.central_widget.splitterMoved.connect(self.new_split)
        self.setCentralWidget(self.central_widget)

    def add_tabs(self):
        was_blocked = self.tabs.blockSignals(True)
        current = self.tabs.currentWidget()
        self.tabs.clear()
        for tab in self.tab_list:
            use_tab = tab['action'].isChecked()
            self.config_store.set('tabs', tab['key'], str(use_tab))
            if tab['object'] and use_tab:
                self.tabs.addTab(tab['object'], tab['name'])
        self.tabs.blockSignals(was_blocked)
        if current:
            self.tabs.setCurrentWidget(current)
        self.new_tab(-1)

    def open_docs(self):
        webbrowser.open_new('http://jim-easterbrook.github.com/Photini/')
    
    def close_files(self):
        self._close_files(False)

    def close_all_files(self):
        self._close_files(True)

    def _close_files(self, all_files):
        if self.image_list.unsaved_files_dialog(all_files=all_files):
            self.image_list.close_files(all_files)

    def closeEvent(self, event):
        self.image_list.unsaved_files_dialog(with_cancel=False)
        self.loggerwindow.shutdown()
        QtGui.QMainWindow.closeEvent(self, event)

    def edit_settings(self):
        dialog = EditSettings(self, self.config_store)
        dialog.exec_()

    @QtCore.pyqtSlot()
    def about(self):
        dialog = QtGui.QMessageBox()
        dialog.setWindowTitle('Photini: about')
        dialog.setText(
            open(os.path.join(data_dir, 'about.html')).read() % (__version__))
        dialog.setDetailedText(
            open(os.path.join(data_dir, 'LICENSE.txt')).read())
        dialog.exec_()

    @QtCore.pyqtSlot(int, int)
    def new_split(self, pos, index):
        self.config_store.set(
            'main_window', 'split', str(self.central_widget.sizes()))

    @QtCore.pyqtSlot(int)
    def new_tab(self, index):
        current = self.tabs.currentWidget()
        if current:
            current.refresh()
            self.image_list.emit_selection()

    @QtCore.pyqtSlot(list)
    def new_selection(self, selection):
        self.close_action.setEnabled(len(selection) > 0)
        self.tabs.currentWidget().new_selection(selection)

    @QtCore.pyqtSlot(bool)
    def new_metadata(self, unsaved_data):
        self.save_action.setEnabled(unsaved_data)

    def resizeEvent(self, event):
        size = self.width(), self.height()
        self.config_store.set('main_window', 'size', str(size))

def main(argv=None):
    if argv:
        sys.argv = argv
    # let PyQt handle its options (need at least one argument after options)
    sys.argv.append('xxx')
    app = QtGui.QApplication(sys.argv)
    del sys.argv[-1]
    # parse remaining arguments
    parser = OptionParser(version='Photini %s' % (__version__),
                          description='Photini photo metadata editor')
    parser.add_option('-v', '--verbose', action='count', default=0,
                      help='increase number of logging messages')
    options, args = parser.parse_args()
    if len(args) != 0:
        parser.error('incorrect number of arguments')
    # create GUI and run application event loop
    main = MainWindow(options.verbose)
    main.show()
    return app.exec_()

if __name__ == "__main__":
    sys.exit(main())



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from datetime import datetime
import os
import subprocess
import sys

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from .flowlayout import FlowLayout
from .metadata import Metadata

class Image(QtGui.QFrame):
    def __init__(self, path, image_list, thumb_size=80, parent=None):
        QtGui.QFrame.__init__(self, parent)
        self.path = os.path.normpath(path)
        self.image_list = image_list
        self.name = os.path.splitext(os.path.basename(self.path))[0]
        self.selected = False
        self.pixmap = None
        self.thumb_size = thumb_size
        self.metadata = Metadata(self.path)
        self.metadata.new_status.connect(self.show_status)
        layout = QtGui.QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(3, 3, 3, 3)
        self.setLayout(layout)
        # label to display image
        self.image = QtGui.QLabel()
        self.image.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        layout.addWidget(self.image, 0, 0, 1, 2)
        # label to display file name
        self.label = QtGui.QLabel(self.name)
        self.label.setAlignment(Qt.AlignRight)
        self.label.setStyleSheet("QLabel { font-size: 12px }")
        layout.addWidget(self.label, 1, 1)
        # label to display status
        self.status = QtGui.QLabel()
        self.status.setAlignment(Qt.AlignLeft)
        self.status.setSizePolicy(
            QtGui.QSizePolicy.Fixed, self.status.sizePolicy().verticalPolicy())
        self.status.setStyleSheet("QLabel { font-size: 12px }")
        self.status.setFont(QtGui.QFont("Dejavu Sans"))
        if not self.status.fontInfo().exactMatch():
            # probably on Windows, try a different font
            self.status.setFont(QtGui.QFont("Segoe UI Symbol"))
        layout.addWidget(self.status, 1, 0)
        self.setFrameStyle(QtGui.QFrame.Panel | QtGui.QFrame.Plain)
        self.setObjectName("thumbnail")
        self.set_selected(False)
        self.show_status(False)
        self._set_thumb_size(self.thumb_size)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.drag_start_pos = event.pos()
        self.image_list.thumb_mouse_press(self.path, event)

    def mouseMoveEvent(self, event):
        if ((event.pos() - self.drag_start_pos).manhattanLength() <
                                    QtGui.QApplication.startDragDistance()):
            return
        drag = QtGui.QDrag(self)
        mimeData = QtCore.QMimeData()
        paths = list()
        for image in self.image_list.get_selected_images():
            paths.append(image.path)
        mimeData.setText(str(paths))
        drag.setMimeData(mimeData)
        dropAction = drag.exec_(Qt.LinkAction)

    def mouseDoubleClickEvent(self, event):
        if sys.platform.startswith('linux'):
            subprocess.call(['xdg-open', self.path])
        elif sys.platform.startswith('darwin'):
            subprocess.call(['open', self.path])
        elif sys.platform.startswith('win'):
            subprocess.call(['start', self.path], shell=True)

    @QtCore.pyqtSlot(bool)
    def show_status(self, changed):
        status = ''
        # set 'geotagged' status
        if not self.metadata.get_item('latlong').empty():
            status += unichr(0x2690)
        # set 'unsaved' status
        if changed:
            status += unichr(0x26A1)
        self.status.setText(status)
        if changed:
            self.image_list.new_metadata.emit(True)

    def _set_thumb_size(self, thumb_size):
        self.thumb_size = thumb_size
        self.image.setFixedSize(self.thumb_size, self.thumb_size)
        margins = self.layout().contentsMargins()
        self.setFixedWidth(self.thumb_size + margins.left() + margins.right() +
                           (self.frameWidth() * 2))

    def set_thumb_size(self, thumb_size):
        self._set_thumb_size(thumb_size)
        self.load_thumbnail()

    def load_thumbnail(self):
        result = False
        if not self.pixmap:
            result = True
            self.pixmap = QtGui.QPixmap(self.path)
            if max(self.pixmap.width(), self.pixmap.height()) > 400:
                # store a scaled down version of image to save memory
                self.pixmap = self.pixmap.scaled(
                    400, 400, Qt.KeepAspectRatio, Qt.SmoothTransformation)
            orientation = self.metadata.get_item('orientation')
            if not orientation.empty() and orientation.value != 1:
                # need to rotate and or reflect image
                transform = QtGui.QTransform()
                if orientation.value in (3, 4):
                    transform = transform.rotate(180.0)
                elif orientation.value in (5, 6):
                    transform = transform.rotate(90.0)
                elif orientation.value in (7, 8):
                    transform = transform.rotate(-90.0)
                if orientation.value in (2, 4, 5, 7):
                    transform = transform.scale(-1.0, 1.0)
                self.pixmap = self.pixmap.transformed(transform)
        self.image.setPixmap(self.pixmap.scaled(
            self.thumb_size, self.thumb_size,
            Qt.KeepAspectRatio, Qt.SmoothTransformation))
        return result

    def set_selected(self, value):
        self.selected = value
        if self.selected:
            self.setStyleSheet("#thumbnail {border: 2px solid red}")
        else:
            self.setStyleSheet("#thumbnail {border: 2px solid grey}")

    def get_selected(self):
        return self.selected

class ScrollArea(QtGui.QScrollArea):
    def __init__(self, parent=None, drop_callback=None):
        QtGui.QScrollArea.__init__(self, parent)
        self.drop_callback = drop_callback
        self.setAcceptDrops(True)

    def dropEvent(self, event):
        if not self.drop_callback:
            return
        file_list = []
        for uri in event.mimeData().urls():
            file_list.append(str(uri.toLocalFile()))
        if not file_list:
            return
        self.drop_callback(file_list)
        event.acceptProposedAction()

    def dragEnterEvent(self, event):
        if self.drop_callback and event.mimeData().hasFormat('text/uri-list'):
            event.acceptProposedAction()

class ImageList(QtGui.QWidget):
    def __init__(self, config_store, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.config_store = config_store
        self.app = QtGui.QApplication.instance()
        self.path_list = list()
        self.image = dict()
        self.last_selected = None
        self.selection_anchor = None
        self.thumb_size = int(self.config_store.get(
            'controls', 'thumb_size', '80'))
        layout = QtGui.QGridLayout()
        layout.setSpacing(0)
        layout.setRowStretch(0, 1)
        layout.setColumnStretch(3, 1)
        self.setLayout(layout)
        layout.setMargin(0)
        # thumbnail display
        self.scroll_area = ScrollArea(drop_callback=self.open_file_list)
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.scroll_area.setWidgetResizable(True)
        layout.addWidget(self.scroll_area, 0, 0, 1, 6)
        self.thumbnails = QtGui.QWidget()
        self.thumbnails.setLayout(FlowLayout(hSpacing=0, vSpacing=0))
        self.scroll_area.setWidget(self.thumbnails)
        QtGui.QShortcut(QtGui.QKeySequence.MoveToPreviousChar,
                        self.scroll_area, self.move_to_prev_thumb)
        QtGui.QShortcut(QtGui.QKeySequence.MoveToNextChar,
                        self.scroll_area, self.move_to_next_thumb)
        QtGui.QShortcut(QtGui.QKeySequence.SelectPreviousChar,
                        self.scroll_area, self.select_prev_thumb)
        QtGui.QShortcut(QtGui.QKeySequence.SelectNextChar,
                        self.scroll_area, self.select_next_thumb)
        QtGui.QShortcut(QtGui.QKeySequence.SelectAll,
                        self.scroll_area, self.select_all)
        # sort key selector
        layout.addWidget(QtGui.QLabel('sort by: '), 1, 0)
        self.sort_name = QtGui.QRadioButton('file name')
        self.sort_name.clicked.connect(self._new_sort_order)
        layout.addWidget(self.sort_name, 1, 1)
        self.sort_date = QtGui.QRadioButton('date taken')
        layout.addWidget(self.sort_date, 1, 2)
        self.sort_date.clicked.connect(self._new_sort_order)
        if eval(self.config_store.get('controls', 'sort_date', 'False')):
            self.sort_date.setChecked(True)
        else:
            self.sort_name.setChecked(True)
        # size selector
        layout.addWidget(QtGui.QLabel('thumbnail size: '), 1, 4)
        self.size_slider = QtGui.QSlider(Qt.Horizontal)
        self.size_slider.setTracking(False)
        self.size_slider.setRange(4, 9)
        self.size_slider.setPageStep(1)
        self.size_slider.setValue(self.thumb_size / 20)
        self.size_slider.setTickPosition(QtGui.QSlider.TicksBelow)
        self.size_slider.setMinimumWidth(140)
        self.size_slider.valueChanged.connect(self._new_thumb_size)
        layout.addWidget(self.size_slider, 1, 5)

    def get_image(self, path):
        if path not in self.path_list:
            return None
        return self.image[path]

    def get_images(self):
        for path in self.path_list:
            yield self.image[path]

    def get_selected_images(self):
        selection = list()
        for path in self.path_list:
            image = self.image[path]
            if image.get_selected():
                selection.append(image)
        return selection

    def mousePressEvent(self, event):
        if self.scroll_area.underMouse():
            self._clear_selection()
            self.last_selected = None
            self.selection_anchor = None
            self.emit_selection()

    @QtCore.pyqtSlot()
    def open_files(self):
        types = []
        if sys.version_info[0] >= 3:
            for ext in QtGui.QImageReader.supportedImageFormats():
                types.append('*.%s' % str(ext, encoding='utf_8'))
        else:
            for ext in QtGui.QImageReader.supportedImageFormats():
                types.append('*.%s' % str(ext))
        path_list = QtGui.QFileDialog.getOpenFileNames(
            self, "Open files", self.config_store.get('paths', 'images', ''),
            "Images (%s);;All files (*)" % ' '.join(types))
        if sys.version_info[0] < 3:
            path_list = map(lambda x: x.toLocal8Bit(), path_list)
        # work around for Qt bug 33992
        # https://bugreports.qt-project.org/browse/QTBUG-33992
        path_list = map(lambda x: str(
            QtCore.QUrl.fromPercentEncoding(x)), path_list)
        if not path_list:
            return
        self.open_file_list(path_list)

    image_list_changed = QtCore.pyqtSignal()
    def open_file_list(self, path_list):
        self.config_store.set(
            'paths', 'images', os.path.dirname(path_list[0]))
        for path in path_list:
            if path in self.path_list:
                continue
            self.path_list.append(path)
            image = Image(path, self, thumb_size=self.thumb_size)
            self.image[path] = image
        self._show_thumbnails()

    def _date_key(self, idx):
        result = self.image[idx].metadata.get_item('date_taken')
        if result.empty():
            result = self.image[idx].metadata.get_item('date_digitised')
        if result.empty():
            result = self.image[idx].metadata.get_item('date_modified')
        if result.empty():
            # use file date as last resort
            result = datetime.fromtimestamp(
                os.path.getmtime(self.image[idx].path))
        else:
            result = result.value
        return result

    sort_order_changed = QtCore.pyqtSignal()
    def _new_sort_order(self):
        self._show_thumbnails()
        self.sort_order_changed.emit()

    def _show_thumbnails(self):
        sort_date = self.sort_date.isChecked()
        self.config_store.set('controls', 'sort_date', str(sort_date))
        if sort_date:
            self.path_list.sort(key=self._date_key)
        else:
            self.path_list.sort()
        layout = self.thumbnails.layout()
        for path in self.path_list:
            image = self.image[path]
            layout.addWidget(image)
            if image.load_thumbnail():
                self.app.processEvents()
                self.scroll_area.ensureWidgetVisible(image)
                self.app.processEvents()
        if self.last_selected:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(self.image[self.last_selected])
        self.image_list_changed.emit()

    def close_files(self, all_files):
        layout = self.thumbnails.layout()
        for path in list(self.path_list):
            image = self.image[path]
            if all_files or image.get_selected():
                self.path_list.remove(path)
                del self.image[path]
                layout.removeWidget(image)
                image.setParent(None)
        self.last_selected = None
        self.selection_anchor = None
        self.emit_selection()
        self.image_list_changed.emit()

    new_metadata = QtCore.pyqtSignal(bool)
    @QtCore.pyqtSlot()
    def save_files(self):
        if_mode = eval(self.config_store.get('files', 'image', 'True'))
        sc_mode = self.config_store.get('files', 'sidecar', 'auto')
        unsaved = False
        for path in list(self.path_list):
            image = self.image[path]
            image.metadata.save(if_mode, sc_mode)
            unsaved = unsaved or image.metadata.changed()
        self.new_metadata.emit(unsaved)

    def unsaved_files_dialog(
            self, all_files=False, with_cancel=True, with_discard=True):
        """Return true if OK to continue with close or quit or whatever"""
        for path in self.path_list:
            image = self.image[path]
            if image.metadata.changed() and (all_files or image.selected):
                break
        else:
            return True
        dialog = QtGui.QMessageBox()
        dialog.setWindowTitle('Photini: unsaved data')
        dialog.setText('<h3>Some images have unsaved metadata.</h3>')
        dialog.setInformativeText('Do you want to save your changes?')
        dialog.setIcon(QtGui.QMessageBox.Warning)
        buttons = QtGui.QMessageBox.Save
        if with_cancel:
            buttons |= QtGui.QMessageBox.Cancel
        if with_discard:
            buttons |= QtGui.QMessageBox.Discard
        dialog.setStandardButtons(buttons)
        dialog.setDefaultButton(QtGui.QMessageBox.Save)
        result = dialog.exec_()
        if result == QtGui.QMessageBox.Save:
            self.save_files()
            return True
        return result == QtGui.QMessageBox.Discard

    def get_selected_images(self):
        selection = list()
        for path in self.path_list:
            image = self.image[path]
            if image.get_selected():
                selection.append(image)
        return selection

    selection_changed = QtCore.pyqtSignal(list)
    def emit_selection(self):
        self.selection_changed.emit(self.get_selected_images())

    def thumb_mouse_press(self, path, event):
        path = str(path)
        if event.modifiers() == Qt.ControlModifier:
            self.select_image(path, multiple_selection=True)
        elif event.modifiers() == Qt.ShiftModifier:
            self.select_image(path, extend_selection=True)
        else:
            self.select_image(path)

    def select_all(self):
        for path in self.path_list:
            image = self.image[path]
            image.set_selected(True)
        self.selection_anchor = None
        self.last_selected = None
        self.emit_selection()

    def move_to_prev_thumb(self):
        self._inc_selection(-1)

    def move_to_next_thumb(self):
        self._inc_selection(1)

    def select_prev_thumb(self):
        self._inc_selection(-1, extend_selection=True)

    def select_next_thumb(self):
        self._inc_selection(1, extend_selection=True)

    def _inc_selection(self, inc, extend_selection=False):
        if self.last_selected:
            idx = self.path_list.index(self.last_selected)
            idx = (idx + inc) % len(self.path_list)
        else:
            idx = 0
        path = self.path_list[idx]
        self.select_image(path, extend_selection=extend_selection)

    @QtCore.pyqtSlot()
    def _new_thumb_size(self):
        self.thumb_size = self.size_slider.value() * 20
        self.config_store.set('controls', 'thumb_size', str(self.thumb_size))
        for path in self.path_list:
            self.image[path].set_thumb_size(self.thumb_size)
        if self.last_selected:
            self.app.processEvents()
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(self.image[self.last_selected])

    def select_image(
            self, path, extend_selection=False, multiple_selection=False):
        image = self.image[path]
        self.scroll_area.ensureWidgetVisible(image)
        if extend_selection and self.selection_anchor:
            self._clear_selection()
            idx1 = self.path_list.index(self.selection_anchor)
            idx2 = self.path_list.index(path)
            for i in range(min(idx1, idx2), max(idx1, idx2) + 1):
                self.image[self.path_list[i]].set_selected(True)
        elif multiple_selection:
            image.set_selected(not image.get_selected())
            self.selection_anchor = path
        else:
            self._clear_selection()
            image.set_selected(True)
            self.selection_anchor = path
        self.last_selected = path
        self.emit_selection()

    def _clear_selection(self):
        for path in self.path_list:
            image = self.image[path]
            if image.get_selected():
                image.set_selected(False)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from ConfigParser import RawConfigParser
import os
import sys

import appdirs
from PyQt4 import QtCore

class ConfigStore(object):
    def __init__(self, name):
        self.config = RawConfigParser()
        self.file_opts = {}
        if sys.version_info[0] >= 3:
            self.file_opts['encoding'] = 'utf-8'
        if hasattr(appdirs, 'user_config_dir'):
            data_dir = appdirs.user_config_dir('photini')
        else:
            data_dir = appdirs.user_data_dir('photini')
        if not os.path.isdir(data_dir):
            os.makedirs(data_dir, mode=0700)
        self.file_name = os.path.join(data_dir, '%s.ini' % name)
        if name == 'editor':
            for old_file_name in (os.path.expanduser('~/photini.ini'),
                                  os.path.join(data_dir, 'photini.ini')):
                if os.path.exists(old_file_name):
                    self.config.read(old_file_name, **self.file_opts)
                    self.save()
                    os.unlink(old_file_name)
        self.config.read(self.file_name, **self.file_opts)
        self.timer = QtCore.QTimer()
        self.timer.setSingleShot(True)
        self.timer.setInterval(3000)
        self.timer.timeout.connect(self.save)
        self.has_section = self.config.has_section

    def get(self, section, option, default=None):
        if self.config.has_option(section, option):
            result = self.config.get(section, option)
            if sys.version_info[0] < 3:
                return result.decode('utf-8')
            return result
        if default is not None:
            self.set(section, option, default)
        return default

    def set(self, section, option, value):
        if not self.config.has_section(section):
            self.config.add_section(section)
        if (self.config.has_option(section, option) and
                self.config.get(section, option) == value):
            return
        if sys.version_info[0] < 3:
            value = value.encode('utf-8')
        self.config.set(section, option, value)
        self.timer.start()

    def remove_section(self, section):
        if not self.config.has_section(section):
            return
        for option in self.config.options(section):
            self.config.remove_option(section, option)
        self.config.remove_section(section)
        self.timer.start()

    def save(self):
        self.config.write(open(self.file_name, 'w', **self.file_opts))
        os.chmod(self.file_name, 0600)



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-13  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import locale
import re
import webbrowser

from PyQt4 import QtGui

from .photinimap import PhotiniMap

class GoogleMap(PhotiniMap):
    def __init__(self, config_store, image_list, parent=None):
        # setting the application name & version stops Google maps
        # using the multitouch interface
        app = QtGui.QApplication.instance()
        app.setApplicationName('chrome')
        app.setApplicationVersion('1.0')
        PhotiniMap.__init__(self, config_store, image_list, parent)

    def load_api(self):
        region = ''
        lang, encoding = locale.getdefaultlocale()
        if lang:
            match = re.match('[a-zA-Z]+[-_]([A-Z]+)', lang)
            if match:
                name = match.group(1)
                if name:
                    region = '&region=%s' % name
        return """
    <script type="text/javascript"
      src="http://maps.googleapis.com/maps/api/js?key=%s&sensor=false%s">
    </script>
""" % ('AIzaSyBPUg_kKGYxyzV0jV7Gg9m4rxme97tE13Y', region)

    def show_terms(self):
        # return a widget to display map terms and conditions
        result = QtGui.QFrame()
        layout = QtGui.QVBoxLayout()
        result.setLayout(layout)
        layout.addWidget(QtGui.QLabel('Search powered by Google'))
        load_tou = QtGui.QPushButton('Terms of Use')
        load_tou.clicked.connect(self.load_tou)
        layout.addWidget(load_tou)
        return result

    def load_tou(self):
        webbrowser.open_new('http://www.google.com/help/terms_maps.html')



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-14  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import webbrowser

from PyQt4 import QtGui

from .photinimap import PhotiniMap

class OpenStreetMap(PhotiniMap):
    def load_api(self):
        return """
    <link rel="stylesheet"
      href="http://cdn.leafletjs.com/leaflet-0.7.3/leaflet.css" />
    <script type="text/javascript">
      var L_NO_TOUCH = true;
    </script>
    <script type="text/javascript"
      src="http://cdn.leafletjs.com/leaflet-0.7.3/leaflet.js">
    </script>
"""

    def show_terms(self):
        # return a widget to display map terms and conditions
        result = QtGui.QFrame()
        layout = QtGui.QVBoxLayout()
        result.setLayout(layout)
        load_tou = QtGui.QPushButton('Search powered by Nominatim')
        load_tou.clicked.connect(self.load_tou_nominatim)
        layout.addWidget(load_tou)
        load_tou = QtGui.QPushButton('Map powered by Leaflet')
        load_tou.clicked.connect(self.load_tou_leaflet)
        layout.addWidget(load_tou)
        load_tou = QtGui.QPushButton('Map data\n©OpenStreetMap contributors')
        load_tou.clicked.connect(self.load_tou_osm)
        layout.addWidget(load_tou)
        load_tou = QtGui.QPushButton('Tiles courtesy of MapQuest')
        load_tou.clicked.connect(self.load_tou_tiles)
        layout.addWidget(load_tou)
        return result

    def load_tou_nominatim(self):
        webbrowser.open_new(
            'http://wiki.openstreetmap.org/wiki/Nominatim_usage_policy')

    def load_tou_leaflet(self):
        webbrowser.open_new('http://leafletjs.com/')

    def load_tou_osm(self):
        webbrowser.open_new('http://www.openstreetmap.org/copyright')

    def load_tou_tiles(self):
        webbrowser.open_new('http://www.mapquest.com/')
